/**
 * Beaver: a compiler front-end compiler
 * Copyright (c) 2003-2012, Alexander Demenchuk
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
package beaver.cc.lexer;

import junit.framework.TestCase;

public class CharRangeTest extends TestCase {

    public void testCharRangeAddAndStringRepr() {
        CharRange range = new CharRange(new CharSpan('0', '9'));
        assertEquals("0-9", range.toString());
        range.add(new CharSpan('A', 'F'));
        assertEquals("0-9A-F", range.toString());
        range.add(new CharSpan('D', 'H'));
        assertEquals("0-9A-H", range.toString());
        range.add(new CharSpan('W', 'Z'));
        assertEquals("0-9A-HW-Z", range.toString());
        range.add(new CharSpan('E', 'X'));
        assertEquals("0-9A-Z", range.toString());
    }

    public void testCharRangeContructorFromString() {
        CharRange range = new CharRange(new CharReader("abct-udefgh-klmnopq-sv-z"));
        assertEquals("a-z", range.toString());
        assertEquals(new CharRange(new CharSpan('a', 'z')), range);

        range = new CharRange(new CharReader("a"));
        assertEquals("a", range.toString());

        range = new CharRange(new CharReader("0-9."));
        assertEquals(".0-9", range.toString());

        range = new CharRange(new CharReader("^0-9"));
        assertEquals("\\x00-/:-\\ufffe", range.toString());

        range = new CharRange(new CharReader("^\\x00-/:-@[-`{-\\ufffe"));
        assertEquals("0-9A-Za-z", range.toString());
    }

    public void testAddChar() {
        CharRange range = new CharRange();
        range.add('a');
        assertEquals("a", range.toString());
        range.add('g');
        assertEquals("ag", range.toString());
        range.add('c');
        assertEquals("acg", range.toString());
        range.add('b');
        assertEquals("a-cg", range.toString());
        range.add('f');
        assertEquals("a-cfg", range.toString());
        range.add('e');
        assertEquals("a-ce-g", range.toString());
        range.add('d');
        assertEquals("a-g", range.toString());
    }

    public void testAccept() {
        CharRange range = new CharRange(new CharReader("abct-udefgh-klmnopq-sv-z"));
        final int[] counter = new int[] { 0 };
        final char[] bounds = new char[] { '\uffff', '\0' };
        range.accept(new CharVisitor() {
            public void visit(char c) {
                counter[0]++;
                if (bounds[0] > c)
                    bounds[0] = c;
                if (bounds[1] < c)
                    bounds[1] = c;
            }
        });
        assertEquals(26, counter[0]);
        assertEquals('a', bounds[0]);
        assertEquals('z', bounds[1]);
    }

    public void testSubtraction_ReduceSpans() {
        CharRange range = new CharRange(new CharReader("a-ps-z"));

        range.sub(new CharRange(new CharReader("l-u")));
        assertEquals("a-kv-z", range.toString());

        range.sub(new CharRange(new CharReader("a-dx-z")));
        assertEquals("e-kvw", range.toString());

        range.sub(new CharRange(new CharReader("a-z")));
        assertEquals("", range.toString());
    }

    public void testSubtraction_RemoveSpans() {
        CharRange range = new CharRange(new CharReader("a-ps-z"));

        range.sub(new CharRange(new CharReader("a-p")));
        assertEquals("s-z", range.toString());

        range.sub(new CharRange(new CharReader("s-z")));
        assertEquals("", range.toString());
    }

    public void testSubtraction_SplitSpans() {
        CharRange range = new CharRange(new CharReader("a-z"));

        range.sub(new CharRange(new CharReader("k-mt-v")));
        assertEquals("a-jn-sw-z", range.toString());
    }
}
