/**
 * Beaver: a compiler front-end compiler
 * Copyright (c) 2003-2012, Alexander Demenchuk
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
package beaver.cc.lexer;

import junit.framework.TestCase;

public class CharSpanTest extends TestCase {

    public void testSize() {
        CharSpan span = new CharSpan('a');
        assertEquals(1, span.size());
        span = new CharSpan('a', 'z');
        assertEquals(26, span.size());
    }

    public void testEqualsCharSpan() {
        CharSpan span1 = new CharSpan('a', 'z');
        CharSpan span2 = new CharSpan('a', 'z');
        assertTrue(span1.equals(span2));
    }

    public void testIntersects() {
        CharSpan span = new CharSpan('c', 'n');
        assertFalse(span.intersects(new CharSpan('a', 'b')));
        assertTrue(span.intersects(new CharSpan('a', 'c')));
        assertTrue(span.intersects(new CharSpan('a', 'f')));
        assertTrue(span.intersects(new CharSpan('k', 'w')));
        assertTrue(span.intersects(new CharSpan('n', 'w')));
        assertFalse(span.intersects(new CharSpan('o', 'w')));
        assertFalse(span.intersects(new CharSpan('r', 'w')));
    }

    public void testIsAdjacentToSpan() {
        CharSpan span = new CharSpan('d', 'n');
        assertFalse(span.isAdjacentTo(new CharSpan('a', 'b')));
        assertTrue(span.isAdjacentTo(new CharSpan('a', 'c')));
        assertFalse(span.isAdjacentTo(new CharSpan('a', 'd')));
        assertFalse(span.isAdjacentTo(new CharSpan('n', 'w')));
        assertTrue(span.isAdjacentTo(new CharSpan('o', 'w')));
        assertFalse(span.isAdjacentTo(new CharSpan('p', 'w')));
    }

    public void testIsSubsetOf() {
        CharSpan span = new CharSpan('c', 'e');
        assertTrue(span.isSubsetOf(new CharSpan('a', 'k')));
        assertTrue(span.isSubsetOf(new CharSpan('c', 'k')));
        assertTrue(span.isSubsetOf(new CharSpan('a', 'f')));
        assertFalse(span.isSubsetOf(new CharSpan('a', 'd')));
        assertFalse(span.isSubsetOf(new CharSpan('d', 'h')));
        assertFalse(span.isSubsetOf(new CharSpan('a', 'b')));
        assertFalse(span.isSubsetOf(new CharSpan('f', 'h')));
    }

    public void testContains() {
        CharSpan span = new CharSpan('c', 'e');
        assertTrue(span.contains('c'));
        assertTrue(span.contains('d'));
        assertTrue(span.contains('e'));
        assertFalse(span.contains('b'));
        assertFalse(span.contains('f'));
    }

    public void testIsAdjacentToChar() {
        CharSpan span = new CharSpan('c', 'e');
        assertTrue(span.isAdjacentTo('b'));
        assertTrue(span.isAdjacentTo('f'));
        assertFalse(span.isAdjacentTo('a'));
        assertFalse(span.isAdjacentTo('c'));
        assertFalse(span.isAdjacentTo('e'));
        assertFalse(span.isAdjacentTo('g'));
    }

    public void testCompareSpan() {
        CharSpan span = new CharSpan('d', 'e');
        assertTrue(span.compare(new CharSpan('g', 'k')) < 0);
        assertTrue(span.compare(new CharSpan('a', 'b')) > 0);
        assertTrue(span.compare(new CharSpan('a', 'c')) == 0);
        assertTrue(span.compare(new CharSpan('a', 'd')) == 0);
        assertTrue(span.compare(new CharSpan('e', 'k')) == 0);
        assertTrue(span.compare(new CharSpan('f', 'k')) == 0);
    }

    public void testCompareChar() {
        CharSpan span = new CharSpan('c', 'e');
        assertTrue(span.compare('a') > 0);
        assertTrue(span.compare('b') == 0);
        assertTrue(span.compare('c') == 0);
        assertTrue(span.compare('d') == 0);
        assertTrue(span.compare('e') == 0);
        assertTrue(span.compare('f') == 0);
        assertTrue(span.compare('g') < 0);
    }

    public void testAddChar() {
        CharSpan span = new CharSpan('c', 'e');
        span.add('b');
        span.add('d');
        span.add('f');
        assertEquals(new CharSpan('b', 'f'), span);
    }

    public void testUnion() {
        CharSpan span = new CharSpan('d', 'l');
        span.add(new CharSpan('k', 'o'));
        assertTrue(span.equals(new CharSpan('d', 'o')));
        span.add(new CharSpan('b', 'k'));
        assertTrue(span.equals(new CharSpan('b', 'o')));
        span.add(new CharSpan('p', 'q'));
        assertTrue(span.equals(new CharSpan('b', 'q')));
        span.add(new CharSpan('a', 'b'));
        assertTrue(span.equals(new CharSpan('a', 'q')));
        span.add(new CharSpan('f', 'n'));
        assertTrue(span.equals(new CharSpan('a', 'q')));
    }

    public void testToString() {
        CharSpan span = new CharSpan('a');
        assertEquals("a", span.toString());
        span.add('b');
        assertEquals("ab", span.toString());
        span.add('c');
        assertEquals("a-c", span.toString());
    }

    public void testAccept() {
        CharSpan span = new CharSpan('a', 'd');
        final int[] counter = new int[] { 0 };
        final char[] bounds = new char[] { '\uffff', '\0' };
        span.accept(new CharVisitor() {
            public void visit(char c) {
                counter[0]++;
                if (bounds[0] > c)
                    bounds[0] = c;
                if (bounds[1] < c)
                    bounds[1] = c;
            }
        });
        assertEquals(4, counter[0]);
        assertEquals('a', bounds[0]);
        assertEquals('d', bounds[1]);
    }

    public void testSubtract_Subset() {
        try {
            CharSpan span = new CharSpan('a', 'z');
            span.sub(new CharSpan('e', 'm'));
            fail("IllegalArgumentException expected");
        } catch (IllegalArgumentException e) {
            assertEquals("e-m is not at the edge of a-z", e.getMessage());
        }
    }

    public void testSubtract_Disjoint() {
        try {
            CharSpan span = new CharSpan('a', 'd');
            span.sub(new CharSpan('e', 'm'));
            fail("IllegalArgumentException expected");
        } catch (IllegalArgumentException e) {
            assertEquals("e-m does not intersect with a-d", e.getMessage());
        }
    }

    public void testSubtract_Left() {
        CharSpan span = new CharSpan('k', 'x');
        span.sub(new CharSpan('a', 'p'));
        assertEquals(new CharSpan('q', 'x'), span);
    }

    public void testSubtract_LeftEdge() {
        CharSpan span = new CharSpan('k', 'x');
        span.sub(new CharSpan('k', 'p'));
        assertEquals(new CharSpan('q', 'x'), span);
    }

    public void testSubtract_LeftChar() {
        CharSpan span = new CharSpan('k', 'x');
        span.sub(new CharSpan('a', 'k'));
        assertEquals(new CharSpan('l', 'x'), span);
    }

    public void testSubtract_Right() {
        CharSpan span = new CharSpan('a', 'x');
        span.sub(new CharSpan('p', 'z'));
        assertEquals(new CharSpan('a', 'o'), span);
    }

    public void testSubtract_RightEdge() {
        CharSpan span = new CharSpan('a', 'x');
        span.sub(new CharSpan('p', 'x'));
        assertEquals(new CharSpan('a', 'o'), span);
    }

    public void testSubtract_RightChar() {
        CharSpan span = new CharSpan('a', 'x');
        span.sub(new CharSpan('x', 'z'));
        assertEquals(new CharSpan('a', 'w'), span);
    }

    public void testSubtractWithSplit_NotAnInnerSubset() {
        try {
            CharSpan span = new CharSpan('a', 'm');
            span.remove(new CharSpan('n', 'w'));
            fail("IllegalArgumentException expected");
        } catch (IllegalArgumentException e) {
            assertEquals("n-w is not a subset of a-m", e.getMessage());
        }

        try {
            CharSpan span = new CharSpan('a', 'm');
            span.remove(new CharSpan('k', 'w'));
            fail("IllegalArgumentException expected");
        } catch (IllegalArgumentException e) {
            assertEquals("k-w is not a subset of a-m", e.getMessage());
        }

        try {
            CharSpan span = new CharSpan('a', 'm');
            span.remove(new CharSpan('k', 'm'));
            fail("IllegalArgumentException expected");
        } catch (IllegalArgumentException e) {
            assertEquals("k-m is not a subset of a-m", e.getMessage());
        }
    }

    public void testSubtractWithSplit() {
        CharSpan span = new CharSpan('a', 'w');
        CharSpan rsub = span.remove(new CharSpan('g', 'm'));
        assertEquals(new CharSpan('a', 'f'), span);
        assertEquals(new CharSpan('n', 'w'), rsub);
    }
}
