/**
 * Beaver: a compiler front-end compiler
 * Copyright (c) 2003-2012, Alexander Demenchuk
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
package beaver.cc.lexer;

import junit.framework.TestCase;

public class DFATest extends TestCase {

    static class Interpreter {
        private DFA    dfa;
        private char[] txt;
        private int    ptr;
        private int    end;

        Interpreter(DFA dfa) {
            this.dfa = dfa;
            this.txt = new char[1024];
        }

        int next(CharReader text) {
            DFA.State st = dfa.start;
            RegExp.Pattern.MatchResult matchResult = null;
            int mark = text.mark();
            end = ptr = 0;
            do {
                if (st.patternMatchResult != null) {
                    mark = text.mark();
                }
                int c = text.readChar();
                if (c < 0) {
                    if (ptr > 0 && matchResult == null)
                        throw new IllegalStateException("unexpected EOF");
                    else
                        break;
                }
                if (st.patternMatchResult != null) {
                    matchResult = st.patternMatchResult;
                    end = ptr;
                }
                DFA.State.Transition tr = st.firstTransition;
                while (tr != null && !tr.onChars.contains((char) c)) {
                    tr = tr.next;
                }
                if (tr == null) {
                    if (matchResult == null)
                        return -1;
                    else
                        break;
                }
                txt[ptr++] = (char) c;
                st = tr.toState;
            } while (st.firstTransition != null);

            if (st.firstTransition == null) {
                if ((matchResult = st.patternMatchResult) == null)
                    throw new IllegalStateException("invalid DFA");
                end = ptr;
            } else {
                text.reset(mark);
            }
            return matchResult != null ? matchResult.id : 0;
        }

        String text() {
            return new String(txt, 0, end);
        }
    }

    public static void print(DFA dfa) {
        for (DFA.State st = dfa.start; st != null; st = st.next) {
            System.out.println(st);
        }
    }

    public void testDecimalScanner() {
        NFA nfa = new NFA(LexerTestFixture.DecimalScanner());
        DFA dfa = new DFA(nfa);
        Interpreter i = new Interpreter(dfa);
        CharReader str = new CharReader("\\t123.456\\r\\n567.89 098.48\\n");
        assertEquals(2, i.next(str));
        assertEquals("\t", i.text());
        assertEquals(1, i.next(str));
        assertEquals("123.456", i.text());
        assertEquals(2, i.next(str));
        assertEquals("\r", i.text());
        assertEquals(2, i.next(str));
        assertEquals("\n", i.text());
        assertEquals(1, i.next(str));
        assertEquals("567.89", i.text());
        assertEquals(2, i.next(str));
        assertEquals(" ", i.text());
        assertEquals(1, i.next(str));
        assertEquals("098.48", i.text());
        assertEquals(2, i.next(str));
        assertEquals("\n", i.text());
        assertEquals(0, i.next(str));
        assertEquals(0, i.next(str));
    }

    public void testIntegerScanner() {
        NFA nfa = new NFA(LexerTestFixture.IntegerScanner());
        DFA dfa = new DFA(nfa);
        Interpreter i = new Interpreter(dfa);
        CharReader str = new CharReader("\t123\n-456 23\n\t0x234\n");
        assertEquals(2, i.next(str));
        assertEquals("\t", i.text());
        assertEquals(1, i.next(str));
        assertEquals("123", i.text());
        assertEquals(2, i.next(str));
        assertEquals("\n", i.text());
        assertEquals(1, i.next(str));
        assertEquals("-456", i.text());
        assertEquals(2, i.next(str));
        assertEquals(" ", i.text());
        assertEquals(1, i.next(str));
        assertEquals("23", i.text());
        assertEquals(2, i.next(str));
        assertEquals("\n", i.text());
        assertEquals(2, i.next(str));
        assertEquals("\t", i.text());
        assertEquals(1, i.next(str));
        assertEquals("0x234", i.text());
        assertEquals(2, i.next(str));
        assertEquals("\n", i.text());
        assertEquals(0, i.next(str));
    }

    public void testIntegerScannerWithBadInput() {
        NFA nfa = new NFA(LexerTestFixture.IntegerScanner());
        DFA dfa = new DFA(nfa);
        Interpreter i = new Interpreter(dfa);
        CharReader str = new CharReader("\ntest\n");
        assertEquals(2, i.next(str));
        assertEquals("\n", i.text());
        assertEquals(-1, i.next(str)); // unrecognizable token
    }
}
