/**
 * Beaver: a compiler front-end compiler
 * Copyright (c) 2003-2012, Alexander Demenchuk
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
package beaver.cc.lexer;

import java.io.File;
import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import java.lang.reflect.Constructor;

import junit.framework.TestCase;
import beaver.CharScanner;
import beaver.TokenInspector;

public class JavaClassCharScannerWriterTest extends TestCase {

    private static final boolean SAVE_SCANNER_CLASS = true;

    private static CharScanner getScanner(final String scannerName, final String superClass, Reader input) throws Exception {
        ClassLoader ldr = new ClassLoader() {
            public Class loadClass(String name) throws ClassNotFoundException {
                if (name.equals(scannerName)) {
                    try {
                        RegExp re = (RegExp) LexerTestFixture.class.getDeclaredMethod(scannerName, null).invoke(null, null);
                        byte[] bc = new JavaClassCharScannerWriter().generate(name, superClass, re);
                        if (SAVE_SCANNER_CLASS) {
                            java.io.FileOutputStream out = new java.io.FileOutputStream(new File(System.getProperty("java.io.tmpdir"), scannerName + ".class"));
                            out.write(bc);
                            out.close();
                        }
                        return defineClass(name, bc, 0, bc.length);
                    } catch (Exception e) {
                        throw new ClassNotFoundException(name, e);
                    }
                }
                return super.loadClass(name);
            }
        };
        Class scannerClass = ldr.loadClass(scannerName);
        Constructor constr = scannerClass.getConstructor(new Class[] { java.io.Reader.class });
        return (CharScanner) constr.newInstance(new Object[] { input });
    }

    static class Term implements beaver.Term {
        int    id;
        int    line;
        int    column;
        Object text;

        Term(int id, int line, int column, Object text) {
            this.id = id;
            this.line = line;
            this.column = column;
            this.text = text;
        }

        public void setId(int newId) {
            id = newId;
        }

        public int getId() {
            return id;
        }

        public Object getText() {
            return text;
        }

        public int getLine() {
            return line;
        }

        public int getColumn() {
            return column;
        }
    }

    public static abstract class TestScanner extends CharScanner {

        protected TestScanner(Reader sourceReader) throws IOException {
            super(sourceReader);
        }

        public beaver.Term makeTerm(int id, int line, int column, Object text) {
            return new Term(id, line, column, text);
        }
    }

    public void testScannerCanBeCreated() throws Exception {
        CharScanner scanner = getScanner("DecimalScanner", "beaver/cc/lexer/JavaClassCharScannerWriterTest$TestScanner", new StringReader("42"));
        assertNotNull(scanner);
    }

    public void testDigitalScanner() throws Exception {
        CharScanner scanner = getScanner("DecimalScanner", "beaver/cc/lexer/JavaClassCharScannerWriterTest$TestScanner", new StringReader("\t123.456\r\n567.89 098.48\n"));
        assertNotNull(scanner);

        beaver.Term t = scanner.getNextToken();
        assertEquals(2, t.getId());
        assertEquals("\t", t.getText());
        t = scanner.getNextToken();
        assertEquals(1, t.getId());
        assertEquals("123.456", t.getText());
        t = scanner.getNextToken();
        assertEquals(2, t.getId());
        assertEquals("\r", t.getText());
        t = scanner.getNextToken();
        assertEquals(2, t.getId());
        assertEquals("\n", t.getText());
        t = scanner.getNextToken();
        assertEquals(1, t.getId());
        assertEquals("567.89", t.getText());
        t = scanner.getNextToken();
        assertEquals(2, t.getId());
        assertEquals(" ", t.getText());
        t = scanner.getNextToken();
        assertEquals(1, t.getId());
        assertEquals("098.48", t.getText());
        t = scanner.getNextToken();
        assertEquals(2, t.getId());
        assertEquals("\n", t.getText());
        t = scanner.getNextToken();
        assertEquals(0, t.getId());
        assertEquals("", t.getText());
        t = scanner.getNextToken();
        assertEquals(0, t.getId());
    }

    public void testIntegerScanner() throws Exception {
        CharScanner scanner = getScanner("IntegerScanner", "beaver/cc/lexer/JavaClassCharScannerWriterTest$TestScanner", new StringReader(" 23 -99 0x89AF 0xabctest"));
        assertNotNull(scanner);

        beaver.Term t = scanner.getNextToken();
        assertEquals(2, t.getId());
        assertEquals(" ", t.getText());
        t = scanner.getNextToken();
        assertEquals(1, t.getId());
        assertEquals("23", t.getText());
        t = scanner.getNextToken();
        assertEquals(2, t.getId());
        assertEquals(" ", t.getText());
        t = scanner.getNextToken();
        assertEquals(1, t.getId());
        assertEquals("-99", t.getText());
        t = scanner.getNextToken();
        assertEquals(2, t.getId());
        assertEquals(" ", t.getText());
        t = scanner.getNextToken();
        assertEquals(1, t.getId());
        assertEquals("0x89AF", t.getText());
        t = scanner.getNextToken();
        assertEquals(2, t.getId());
        assertEquals(" ", t.getText());
        t = scanner.getNextToken();
        assertEquals(1, t.getId());
        assertEquals("0xabc", t.getText());
        t = scanner.getNextToken();
        assertEquals(-1, t.getId()); // unrecognizable token
        assertEquals("t", t.getText());
        t = scanner.getNextToken();
        assertEquals(-1, t.getId());
        assertEquals("e", t.getText());
        t = scanner.getNextToken();
        assertEquals(-1, t.getId());
        assertEquals("s", t.getText());
        t = scanner.getNextToken();
        assertEquals(-1, t.getId());
        assertEquals("t", t.getText());
        t = scanner.getNextToken();
        assertEquals(0, t.getId());
    }

    public void testLogicCalculatorScanner() throws Exception {
        CharScanner scanner = getScanner("LogicCalculatorScanner", "beaver/cc/lexer/JavaClassCharScannerWriterTest$TestScanner", new StringReader(
            "cat android and oracle\n\tor not note\n\ndone"));
        assertNotNull(scanner);

        beaver.Term t = scanner.getNextToken();
        assertEquals(5, t.getId());
        assertEquals("cat", t.getText());
        t = scanner.getNextToken();
        assertEquals(6, t.getId());
        assertEquals(" ", t.getText());
        t = scanner.getNextToken();
        assertEquals(5, t.getId());
        assertEquals("android", t.getText());
        t = scanner.getNextToken();
        assertEquals(6, t.getId());
        assertEquals(" ", t.getText());
        t = scanner.getNextToken();
        assertEquals(2, t.getId());
        assertEquals("and", t.getText());
        t = scanner.getNextToken();
        assertEquals(6, t.getId());
        assertEquals(" ", t.getText());
        t = scanner.getNextToken();
        assertEquals(5, t.getId());
        assertEquals("oracle", t.getText());
        t = scanner.getNextToken();
        assertEquals(1, t.getId());
        assertEquals("\n", t.getText());
        t = scanner.getNextToken();
        assertEquals(6, t.getId());
        assertEquals("\t", t.getText());
        t = scanner.getNextToken();
        assertEquals(3, t.getId());
        assertEquals("or", t.getText());
        t = scanner.getNextToken();
        assertEquals(6, t.getId());
        assertEquals(" ", t.getText());
        t = scanner.getNextToken();
        assertEquals(4, t.getId());
        assertEquals("not", t.getText());
        t = scanner.getNextToken();
        assertEquals(6, t.getId());
        assertEquals(" ", t.getText());
        t = scanner.getNextToken();
        assertEquals(5, t.getId());
        assertEquals("note", t.getText());
        t = scanner.getNextToken();
        assertEquals(1, t.getId());
        assertEquals("\n", t.getText());
        t = scanner.getNextToken();
        assertEquals(1, t.getId());
        assertEquals("\n", t.getText());
        t = scanner.getNextToken();
        assertEquals(5, t.getId());
        assertEquals("done", t.getText());
        t = scanner.getNextToken();
        assertEquals(0, t.getId());
    }

    public void testEventfulLogicCalculatorScanner() throws Exception {
        String textToScan = "cat android and oracle\n\tor not note\n\ndone";
        CharScanner scanner = getScanner("EventfulLogicCalculatorScanner", "beaver/cc/lexer/JavaClassCharScannerWriterTest$TestScanner", new StringReader(textToScan));
        assertNotNull(scanner);

        beaver.Term t = scanner.getNextToken();
        assertEquals(5, t.getId());
        assertEquals("cat", t.getText());
        assertEquals(1, t.getLine());
        assertEquals(1, t.getColumn());
        t = scanner.getNextToken();
        assertEquals(5, t.getId());
        assertEquals("android", t.getText());
        assertEquals(1, t.getLine());
        assertEquals(5, t.getColumn());
        t = scanner.getNextToken();
        assertEquals(2, t.getId());
        assertEquals("and", t.getText());
        assertEquals(1, t.getLine());
        assertEquals(13, t.getColumn());
        t = scanner.getNextToken();
        assertEquals(5, t.getId());
        assertEquals("oracle", t.getText());
        assertEquals(1, t.getLine());
        assertEquals(17, t.getColumn());
        t = scanner.getNextToken();
        assertEquals(3, t.getId());
        assertEquals("or", t.getText());
        assertEquals(2, t.getLine());
        assertEquals(2, t.getColumn());
        t = scanner.getNextToken();
        assertEquals(4, t.getId());
        assertEquals("not", t.getText());
        assertEquals(2, t.getLine());
        assertEquals(5, t.getColumn());
        t = scanner.getNextToken();
        assertEquals(5, t.getId());
        assertEquals("note", t.getText());
        assertEquals(2, t.getLine());
        assertEquals(9, t.getColumn());
        t = scanner.getNextToken();
        assertEquals(5, t.getId());
        assertEquals("done", t.getText());
        assertEquals(4, t.getLine());
        assertEquals(1, t.getColumn());
        t = scanner.getNextToken();
        assertEquals(0, t.getId());
    }

    public static abstract class InspectingScannerBase extends CharScanner {

        protected InspectingScannerBase(Reader input) throws IOException {
            super(input);
        }

        void setInspector(TokenInspector insp) {
            super.inspector = insp;
        }
    }

    public void testInspectingLogicCalculatorScanner() throws Exception {
        String textToScan = "cat android and oracle\n\tor not note\n\ndone";
        CharScanner scanner = getScanner("EventfulLogicCalculatorScanner", "beaver/cc/lexer/JavaClassCharScannerWriterTest$TestScanner", new StringReader(textToScan));
        assertNotNull(scanner);

        beaver.Term t = scanner.getNextToken();
        assertEquals(5, t.getId());
        assertEquals("cat", t.getText());
        t = scanner.getNextToken();
        assertEquals(5, t.getId());
        assertEquals("android", t.getText());
        t = scanner.getNextToken();
        assertEquals(2, t.getId());
        assertEquals("and", t.getText());
        t = scanner.getNextToken();
        assertEquals(5, t.getId());
        assertEquals("oracle", t.getText());
        t = scanner.getNextToken();
        assertEquals(3, t.getId());
        assertEquals("or", t.getText());
        t = scanner.getNextToken();
        assertEquals(4, t.getId());
        assertEquals("not", t.getText());

        scanner.setTokenInspector(new TokenInspector() {
            public boolean isValid(int tokenId) {
                return tokenId != 5;
            }
            public boolean isKeyword(int tokenId) {
                return tokenId != 5;
            }
        });

        t = scanner.getNextToken();
        assertEquals(4, t.getId());
        assertNull(t.getText()); // a keyword

        scanner.setTokenInspector(null);

        t = scanner.getNextToken();
        assertEquals(5, t.getId());
        assertEquals("e", t.getText());
        t = scanner.getNextToken();
        assertEquals(5, t.getId());
        assertEquals("done", t.getText());
        t = scanner.getNextToken();
        assertEquals(0, t.getId());
    }

    public void testRangeScanner() throws Exception {
        String text = "    dot = any \\ [\\n\\r]\r\n;\r\n";
        CharScanner scanner = getScanner("RangeScanner", "beaver/cc/Scanner", new StringReader(text));
        // ; = \ NAME RANGE
        beaver.Term t = scanner.getNextToken();
        assertEquals(4, t.getId());
        assertEquals("dot", t.getText());
        t = scanner.getNextToken();
        assertEquals(2, t.getId());
        t = scanner.getNextToken();
        assertEquals(4, t.getId());
        assertEquals("any", t.getText());
        t = scanner.getNextToken();
        assertEquals(3, t.getId());
        t = scanner.getNextToken();
        assertEquals(5, t.getId());
        assertEquals("[\\n\\r]", t.getText());
        t = scanner.getNextToken();
        assertEquals(1, t.getId());
        t = scanner.getNextToken();
        assertEquals(0, t.getId());
    }
}
