/**
 * Beaver: a compiler front-end compiler
 * Copyright (c) 2003-2012, Alexander Demenchuk
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
package beaver.cc.lexer;

/**
 * LexerTestFixture is a utility that builds different regular expressions for unit tests.
 */
class LexerTestFixture {
    /**
     * DigitalScanner matches digits and white-space
     *
     * @return regular expression for a scanner
     */
    static RegExp DigitalScanner() {
        return
            new RegExp.Alt(
                new RegExp.Pattern(
                    new RegExp.MatchRange("0-9"),
                    new RegExp.Pattern.MatchResult(1, 9999)),
                new RegExp.Pattern(
                    new RegExp.MatchRange(" \\t\\n\\r"),
                    new RegExp.Pattern.MatchResult(2, 9998))
            );
    }

    /**
     * DecimalScanner recognizes decimal representation of numbers with fractions
     *
     * @return regular expression for a scanner
     */
    static RegExp DecimalScanner() {
        return
            new RegExp.Alt(
                new RegExp.Pattern(
                    new RegExp.Cat(
                        new RegExp.Quantified(
                            new RegExp.MatchRange("0-9"),
                            '+'),
                        new RegExp.Quantified(
                            new RegExp.Cat(
                                new RegExp.MatchChar('.'),
                                new RegExp.Quantified(
                                    new RegExp.MatchRange("0-9"),
                                    '+')),
                            '?')),
                    new RegExp.Pattern.MatchResult(1, 9999)),
                new RegExp.Pattern(
                    new RegExp.MatchRange(" \\t\\n\\r"),
                    new RegExp.Pattern.MatchResult(2, 9998))
            );
    }

    /**
     * IntegerScanner recognizes whole numbers expressed as possibly negative decimals without
     * leading zero or hexadecimals that start with 0x
     *
     * @return regular expression for a scanner
     */
    static RegExp IntegerScanner() {
        return
            new RegExp.Alt(
                new RegExp.Pattern(
                    new RegExp.Alt(
                        new RegExp.Cat(
                            new RegExp.Quantified(
                                new RegExp.MatchChar('-'),
                                '?'),
                            new RegExp.Cat(
                                new RegExp.MatchRange("1-9"),
                                new RegExp.Quantified(
                                    new RegExp.MatchRange("0-9"),
                                    '*'))),
                        new RegExp.Cat(
                            new RegExp.Cat(
                                new RegExp.MatchChar('0'),
                                new RegExp.MatchChar('x')),
                            new RegExp.Quantified(
                                new RegExp.MatchRange("0-9A-Fa-f"),
                                '+'))),
                    new RegExp.Pattern.MatchResult(1, 9999)),
                new RegExp.Pattern(
                    new RegExp.MatchRange(" \\t\\n\\r"),
                    new RegExp.Pattern.MatchResult(2, 9998))
            );
    }

    static RegExp LogicCalculatorScanner() {
        return
        new RegExp.Alt(
            new RegExp.Pattern(
                new RegExp.Alt(
                    RegExp.matchText("\r"),
                    new RegExp.Alt(
                        RegExp.matchText("\n"),
                        RegExp.matchText("\r\n"))),
                new RegExp.Pattern.MatchResult(1, 9999)),
        new RegExp.Alt(
            new RegExp.Pattern(
                RegExp.matchText("and"),
                new RegExp.Pattern.MatchResult(2, 9998)),
        new RegExp.Alt(
            new RegExp.Pattern(
                RegExp.matchText("or"),
                new RegExp.Pattern.MatchResult(3, 9997)),
        new RegExp.Alt(
            new RegExp.Pattern(
                RegExp.matchText("not"),
                new RegExp.Pattern.MatchResult(4, 9996)),
        new RegExp.Alt(
            new RegExp.Pattern(
                new RegExp.Cat(
                    new RegExp.MatchRange("a-zA-Z_"),
                    new RegExp.Quantified(
                        new RegExp.MatchRange("0-9a-zA-Z_"), '*')
                    ),
                new RegExp.Pattern.MatchResult(5, 9995)),
            new RegExp.Pattern(
                new RegExp.MatchRange(" \t"),
                new RegExp.Pattern.MatchResult(6, 9994))
        )))));
    }

    static RegExp EventfulLogicCalculatorScanner() {
        return
        new RegExp.Alt(
            new RegExp.Pattern(
                new RegExp.Alt(
                    RegExp.matchText("\r"),
                    new RegExp.Alt(
                        RegExp.matchText("\n"),
                        RegExp.matchText("\r\n"))),
                new RegExp.Pattern.MatchResult(1, 9999, new String[] {"NewLine", "Skip"})),
        new RegExp.Alt(
            new RegExp.Pattern(
                RegExp.matchText("and"),
                new RegExp.Pattern.MatchResult(2, 9998)),
        new RegExp.Alt(
            new RegExp.Pattern(
                RegExp.matchText("or"),
                new RegExp.Pattern.MatchResult(3, 9997)),
        new RegExp.Alt(
            new RegExp.Pattern(
                RegExp.matchText("not"),
                new RegExp.Pattern.MatchResult(4, 9996)),
        new RegExp.Alt(
            new RegExp.Pattern(
                new RegExp.Cat(
                    new RegExp.MatchRange("a-zA-Z_"),
                    new RegExp.Quantified(
                        new RegExp.MatchRange("0-9a-zA-Z_"), '*')
                    ),
                new RegExp.Pattern.MatchResult(5, 9995)),
            new RegExp.Pattern(
                new RegExp.MatchRange(" \t"),
                new RegExp.Pattern.MatchResult(6, 9994, new String[] {"Skip"}))
        )))));
    }

    static RegExp RangeScanner() {
        int prec = Integer.MAX_VALUE;
        RegExp.Alt rules =
            new RegExp.Alt(
                new RegExp.Pattern(
                    new RegExp.Alt(
                        RegExp.matchText("\r"),
                        RegExp.matchText("\n")
                    ).add(
                        RegExp.matchText("\r\n")
                    ),
                    new RegExp.Pattern.MatchResult(-1, prec--, new String[] {"NewLine", "Skip"})
                ),
                new RegExp.Pattern(
                    new RegExp.MatchRange(" \t"),
                    new RegExp.Pattern.MatchResult(-1, prec--, new String[] {"Skip"})
                )
            );
        rules.add(
                new RegExp.Pattern(
                    RegExp.matchText(";"),
                    new RegExp.Pattern.MatchResult(1, prec--)
                )
            );
        rules.add(
                new RegExp.Pattern(
                    RegExp.matchText("="),
                    new RegExp.Pattern.MatchResult(2, prec--)
                )
            );
        rules.add(
                new RegExp.Pattern(
                    RegExp.matchText("\\"),
                    new RegExp.Pattern.MatchResult(3, prec--)
                )
            );
        RegExp.MatchRange any    = new RegExp.MatchRange("^");
        RegExp.MatchRange dot    = any.subtract("\\n\\r");
        RegExp.MatchRange esc    = dot.subtract("\\\\");
        RegExp.MatchRange letter = new RegExp.MatchRange("A-Za-z_");
        RegExp.MatchRange digit  = new RegExp.MatchRange("0-9");
        rules.add(
                new RegExp.Pattern(
                        new RegExp.Cat(
                            letter,
                            new RegExp.Quantified(
                                new RegExp.Alt(
                                    letter,
                                    digit
                                ),
                                '*'
                            )
                        ),
                        new RegExp.Pattern.MatchResult(4, prec--)
                    )
            ).add(
                new RegExp.Pattern(
                        new RegExp.Cat(
                            RegExp.matchText("["),
                            new RegExp.Cat(
                                new RegExp.Quantified(
                                    RegExp.matchText("^"),
                                    '?'
                                ),
                                new RegExp.Cat(
                                    new RegExp.Quantified(
                                        new RegExp.Alt(
                                            esc.subtract("]"),
                                            new RegExp.Cat (
                                                RegExp.matchText("\\"),
                                                dot
                                            )
                                        ),
                                        '*'
                                    ),
                                    RegExp.matchText("]")
                                )
                            )
                        ),
                        new RegExp.Pattern.MatchResult(5, prec--)
                    )
                );
        return rules;
    }
}
