/**
 * Beaver: a compiler front-end compiler
 * Copyright (c) 2003-2012, Alexander Demenchuk
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
package beaver.cc.parser;

import junit.framework.TestCase;

/**
 * Eval = StmtList
 *      ;
 * StmtList
 *      = Stmt
 *      | StmtList Stmt
 *      ;
 * Stmt = { Print } "print" Expr ";"
 *      | { Set   } ID "=" Expr ";"
 *      ;
 *
 * Expr = { Number } NUM
 *      | { Var }    ID
 *      | { Nested } "(" Expr ")"
 *      | { Add }    leftTerm:Expr   "+" rightTerm:Expr
 *      | { Sub }    minuend:Expr    "-" subtrahend:Expr
 *      | { Mul }    leftFactor:Expr "*" rightFactor:Expr
 *      | { Div }    dividend:Expr   "/" divisor:Expr
 *      ;
 */
public class GrammarTest extends TestCase {

    Grammar grammar;

    public GrammarTest(String name) {
        super(name);

        Symbol.Terminal print    = new Symbol.Terminal("\"print\"", "print");
        Symbol.Terminal eq       = new Symbol.Terminal("\"=\"", "=");
        Symbol.Terminal semicol  = new Symbol.Terminal("\";\"", ";");
        Symbol.Terminal lp       = new Symbol.Terminal("\"(\"", "(");
        Symbol.Terminal rp       = new Symbol.Terminal("\")\"", ")");
        Symbol.Terminal plus     = new Symbol.Terminal("\"+\"", "+");
        Symbol.Terminal minus    = new Symbol.Terminal("\"-\"", "-");
        Symbol.Terminal asterisk = new Symbol.Terminal("\"*\"", "*");
        Symbol.Terminal slash    = new Symbol.Terminal("\"/\"", "/");
        Symbol.Terminal id       = new Symbol.Terminal("ID");
        Symbol.Terminal num      = new Symbol.Terminal("NUM");

        Symbol.NonTerminal eval     = new Symbol.NonTerminal("Eval");
        Symbol.NonTerminal stmtlist = new Symbol.NonTerminal("StmtList");
        Symbol.NonTerminal stmt     = new Symbol.NonTerminal("Stmt");
        Symbol.NonTerminal expr     = new Symbol.NonTerminal("Expr");

        Production[] rules = new Production[12];
        rules[ 0] = new Production(eval, new Production.RHSElement[] { new Production.RHSElement(stmtlist) });
        rules[ 1] = new Production(stmtlist, new Production.RHSElement[] { new Production.RHSElement(stmt) });
        rules[ 2] = new Production(stmtlist, new Production.RHSElement[] { new Production.RHSElement(stmtlist), new Production.RHSElement(stmt) });
        rules[ 3] = new Production("Print",  stmt, new Production.RHSElement[] { new Production.RHSElement(print), new Production.RHSElement(expr), new Production.RHSElement(semicol) });
        rules[ 4] = new Production("Set",    stmt, new Production.RHSElement[] { new Production.RHSElement(id), new Production.RHSElement(eq), new Production.RHSElement(expr), new Production.RHSElement(semicol) });
        rules[ 5] = new Production("Number", expr, new Production.RHSElement[] { new Production.RHSElement(num) });
        rules[ 6] = new Production("Var",    expr, new Production.RHSElement[] { new Production.RHSElement(id) });
        rules[ 7] = new Production("Nested", expr, new Production.RHSElement[] { new Production.RHSElement(lp), new Production.RHSElement(expr), new Production.RHSElement(rp) });
        rules[ 8] = new Production("Add",    expr, new Production.RHSElement[] { new Production.RHSElement("leftTerm", expr), new Production.RHSElement(plus), new Production.RHSElement("rightTerm", expr) });
        rules[ 9] = new Production("Sub",    expr, new Production.RHSElement[] { new Production.RHSElement("minuend", expr), new Production.RHSElement(minus), new Production.RHSElement("subtrahend", expr) });
        rules[10] = new Production("Mul",    expr, new Production.RHSElement[] { new Production.RHSElement("leftFactor", expr), new Production.RHSElement(asterisk), new Production.RHSElement("rightFactor", expr) });
        rules[11] = new Production("Div",    expr, new Production.RHSElement[] { new Production.RHSElement("dividend", expr), new Production.RHSElement(slash), new Production.RHSElement("divisor", expr) });

        assignProductionsToNonterminals(rules);

        plus.setPrecedence('\u0001', 'L');
        minus.setPrecedence('\u0001', 'L');
        asterisk.setPrecedence('\u0002', 'L');
        slash.setPrecedence('\u0002', 'L');

        grammar = new Grammar(rules);
    }

    static void assignProductionsToNonterminals(Production[] rules) {
        for (int i = 0; i < rules.length; i++) {
            if (rules[i].lhs.rules == null) {
                int n = 1;
                for (int j = i + 1; j < rules.length; j++) {
                    if (rules[j].lhs == rules[i].lhs) {
                        n++;
                    }
                }
                Production[] ntRules = new Production[n];
                ntRules[0] = rules[i];
                for (int j = i + 1; j < rules.length; j++) {
                    if (rules[j].lhs == rules[i].lhs) {
                        ntRules[j - i] = rules[j];
                    }
                }
                rules[i].lhs.setRules(ntRules);
            }
        }
    }

    public void testGoal() {
        Symbol.NonTerminal goal = grammar.getGoal();
        assertTrue("Goal is a nonterminal", goal instanceof Symbol.NonTerminal);
        assertEquals("Eval", goal.name);
        assertNotNull("Goal is delegated", goal.delegate);
        assertEquals("StmtList", goal.delegate.name);
        assertEquals("Goal reported name", "StmtList", goal.getName());
    }

    public void testError() {
        assertNull("This grammar does not use ERROR", grammar.getError());
    }

    public void testLastTerminalId() {
        assertEquals("ID of the last Terminal", 11, grammar.getLastTerminalId());
    }

    public void testLastKeyword() {
        Symbol.Terminal kwd = grammar.getLastKeyword();
        assertEquals("ID of the last keyword", 9, kwd.getId());
    }
}
