/**
 * Beaver: a compiler front-end compiler
 * Copyright (c) 2003-2012, Alexander Demenchuk
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
package beaver.cc.parser;

import java.io.IOException;
import java.lang.reflect.Constructor;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Map;
import java.util.TreeMap;

import junit.framework.TestCase;
import beaver.Parser;
import beaver.Scanner;
import beaver.TokenInspector;

public class JavaClassParserWriterTest extends TestCase {

    // Generated AST:
	public static abstract class Node {
		Node parent;
		abstract void accept(Walker walker);
		abstract Node dup();
		static String asString(Object object) {
			return object == null ? "" : " " + object.toString();
		}
	}
	public static class Term extends Node implements beaver.Term {
		static final String[] keywords = { "'<EOF>'", "'print'", "';'", "'='", "'('", "')'", "'+'", "'-'", "'*'", "'/'" };
		int    id;
		int    line;
		int    column;
		Object text;
		public Term(int id, int line, int column, Object text) {
			this.id = id;
			this.line = line;
			this.column = column;
			this.text = text;
		}
		private Term(Term term) {
			this.id = term.id;
			this.line = term.line;
			this.column = term.column;
			this.text = term.text;
		}
		public void setId(int newId) {
			id = newId;
		}
		public int getId() {
			return id;
		}
		public Object getText() {
			return text;
		}
		public int getLine() {
			return line;
		}
		public int getColumn() {
			return column;
		}
		void accept(Walker walker) {
			walker.visit(this);
		}
		Node dup() {
			return new Term(this);
		}
		public String toString() {
			return (id < keywords.length ? keywords[id] : text.toString());
		}
	}
	public static abstract class NodeList extends Node {
		static class Elem {
			Elem next;
			Node node;
			Elem (Node node) {
				this.node = node;
			}
		}
		Elem first;
		Elem last;
		int  size;
		NodeList() {}
		NodeList(Node node) {
			node.parent = this;
			last = first = new Elem(node);
			size = 1;
		}
		protected NodeList(NodeList list) {
			Elem elem = list.first;
			if (elem != null) {
				(last = first = new Elem(elem.node.dup())).node.parent = this;
				size = 1;
				for (elem = elem.next; elem != null; elem = elem.next) {
					add(elem.node.dup());
				}
			}
		}
		public void add(Node node) {
			node.parent = this;
			last = last.next = new Elem(node);
			size++;
		}
		public void remove(Node node) {
			Elem elem = first, prev = null;
			while (elem != null && elem.node != node) {
				elem = (prev = elem).next;
			}
			if (elem == null) {
				throw new IllegalArgumentException("not found");
			}
			if (prev != null) {
				prev.next = elem.next;
			} else {
				first = elem.next;
			}
			size--;
		}
		public void replace(Node node, Node newNode) {
			Elem elem = first;
			while (elem != null && elem.node != node) {
				elem = elem.next;
			}
			if (elem == null) {
				throw new IllegalArgumentException("not found");
			}
			(elem.node = newNode).parent = this;
		}
		public void replace(Node node, NodeList list) {
			Elem elem = first, prev = null;
			while (elem != null && elem.node != node) {
				elem = (prev = elem).next;
			}
			if (elem == null) {
				throw new IllegalArgumentException("not found");
			}
			for (Elem e = list.first; e != null; e = e.next) {
				e.node.parent = this;
			}
			if (prev != null) {
				prev.next = list.first;
			} else {
				first = list.first;
			}
			list.last.next = elem.next;
			size += list.size - 1;
			list.first = list.last = null;
			list.size = 0;
		}
		public String toString() {
			if (first != null) {
				String str = first.node.toString();
				for (Elem e = first.next; e != null; e = e.next) {
					str += ", " + e.node;
				}
				return str;
			}
			return "";
		}
	}
	public static class StmtList extends NodeList {
		private StmtList(StmtList list) {
			super(list);
		}
		public StmtList() {
			super();
		}
		public StmtList(Stmt elem) {
			super(elem);
		}
		Node dup() {
			return new StmtList(this);
		}
		void accept(Walker walker) {
			walker.visit(this);
		}
		public String toString() {
			return "(StmtList " + super.toString() + ")";
		}
	}
	public static abstract class Stmt extends Node {
		public static class Print extends Stmt {
			Expr expr;
			public Print(Expr expr) {
				(this.expr = expr).parent = this;
			}
			private Print(Print node) {
				(this.expr = (Expr) node.expr.dup()).parent = this;
			}
			Node dup() {
				return new Print(this);
			}
			void accept(Walker walker) {
				walker.visit(this);
			}
			public String toString() {
				return "(Stmt.Print" + " " + expr + ")";
			}
		}
		public static class Set extends Stmt {
			Term id;
			Expr expr;
			public Set(Term id, Expr expr) {
				(this.id = id).parent = this;
				(this.expr = expr).parent = this;
			}
			private Set(Set node) {
				(this.id = (Term) node.id.dup()).parent = this;
				(this.expr = (Expr) node.expr.dup()).parent = this;
			}
			Node dup() {
				return new Set(this);
			}
			void accept(Walker walker) {
				walker.visit(this);
			}
			public String toString() {
				return "(Stmt.Set" + " " + id + " " + expr + ")";
			}
		}
	}
	public static abstract class Expr extends Node {
		public static class Number extends Expr {
			Term num;
			public Number(Term num) {
				(this.num = num).parent = this;
			}
			private Number(Number node) {
				(this.num = (Term) node.num.dup()).parent = this;
			}
			Node dup() {
				return new Number(this);
			}
			void accept(Walker walker) {
				walker.visit(this);
			}
			public String toString() {
				return "(Expr.Number" + " " + num + ")";
			}
		}
		public static class Var extends Expr {
			Term id;
			public Var(Term id) {
				(this.id = id).parent = this;
			}
			private Var(Var node) {
				(this.id = (Term) node.id.dup()).parent = this;
			}
			Node dup() {
				return new Var(this);
			}
			void accept(Walker walker) {
				walker.visit(this);
			}
			public String toString() {
				return "(Expr.Var" + " " + id + ")";
			}
		}
		public static class Nested extends Expr {
			Expr expr;
			public Nested(Expr expr) {
				(this.expr = expr).parent = this;
			}
			private Nested(Nested node) {
				(this.expr = (Expr) node.expr.dup()).parent = this;
			}
			Node dup() {
				return new Nested(this);
			}
			void accept(Walker walker) {
				walker.visit(this);
			}
			public String toString() {
				return "(Expr.Nested" + " " + expr + ")";
			}
		}
		public static class Add extends Expr {
			Expr leftTerm;
			Expr rightTerm;
			public Add(Expr leftTerm, Expr rightTerm) {
				(this.leftTerm = leftTerm).parent = this;
				(this.rightTerm = rightTerm).parent = this;
			}
			private Add(Add node) {
				(this.leftTerm = (Expr) node.leftTerm.dup()).parent = this;
				(this.rightTerm = (Expr) node.rightTerm.dup()).parent = this;
			}
			Node dup() {
				return new Add(this);
			}
			void accept(Walker walker) {
				walker.visit(this);
			}
			public String toString() {
				return "(Expr.Add" + " " + leftTerm + " " + rightTerm + ")";
			}
		}
		public static class Sub extends Expr {
			Expr minuend;
			Expr subtrahend;
			public Sub(Expr minuend, Expr subtrahend) {
				(this.minuend = minuend).parent = this;
				(this.subtrahend = subtrahend).parent = this;
			}
			private Sub(Sub node) {
				(this.minuend = (Expr) node.minuend.dup()).parent = this;
				(this.subtrahend = (Expr) node.subtrahend.dup()).parent = this;
			}
			Node dup() {
				return new Sub(this);
			}
			void accept(Walker walker) {
				walker.visit(this);
			}
			public String toString() {
				return "(Expr.Sub" + " " + minuend + " " + subtrahend + ")";
			}
		}
		public static class Mul extends Expr {
			Expr leftFactor;
			Expr rightFactor;
			public Mul(Expr leftFactor, Expr rightFactor) {
				(this.leftFactor = leftFactor).parent = this;
				(this.rightFactor = rightFactor).parent = this;
			}
			private Mul(Mul node) {
				(this.leftFactor = (Expr) node.leftFactor.dup()).parent = this;
				(this.rightFactor = (Expr) node.rightFactor.dup()).parent = this;
			}
			Node dup() {
				return new Mul(this);
			}
			void accept(Walker walker) {
				walker.visit(this);
			}
			public String toString() {
				return "(Expr.Mul" + " " + leftFactor + " " + rightFactor + ")";
			}
		}
		public static class Div extends Expr {
			Expr dividend;
			Expr divisor;
			public Div(Expr dividend, Expr divisor) {
				(this.dividend = dividend).parent = this;
				(this.divisor = divisor).parent = this;
			}
			private Div(Div node) {
				(this.dividend = (Expr) node.dividend.dup()).parent = this;
				(this.divisor = (Expr) node.divisor.dup()).parent = this;
			}
			Node dup() {
				return new Div(this);
			}
			void accept(Walker walker) {
				walker.visit(this);
			}
			public String toString() {
				return "(Expr.Div" + " " + dividend + " " + divisor + ")";
			}
		}
	}
	public static class Walker {
		void visit(Term node) {}
		boolean enter(StmtList node) {
			return true;
		}
		void leave(StmtList node) {}
		void visit(StmtList node) {
			if (enter(node)) {
				for (NodeList.Elem e = node.first; e != null; e = e.next) {
					e.node.accept(this);
				}
				leave(node);
			}
		}
		boolean enter(Stmt.Print node) {
			return true;
		}
		void leave(Stmt.Print node) {}
		void visit(Stmt.Print node) {
			if (enter(node)) {
				node.expr.accept(this);
				leave(node);
			}
		}
		boolean enter(Stmt.Set node) {
			return true;
		}
		void leave(Stmt.Set node) {}
		void visit(Stmt.Set node) {
			if (enter(node)) {
				visit(node.id);
				node.expr.accept(this);
				leave(node);
			}
		}
		boolean enter(Expr.Number node) {
			return true;
		}
		void leave(Expr.Number node) {}
		void visit(Expr.Number node) {
			if (enter(node)) {
				visit(node.num);
				leave(node);
			}
		}
		boolean enter(Expr.Var node) {
			return true;
		}
		void leave(Expr.Var node) {}
		void visit(Expr.Var node) {
			if (enter(node)) {
				visit(node.id);
				leave(node);
			}
		}
		boolean enter(Expr.Nested node) {
			return true;
		}
		void leave(Expr.Nested node) {}
		void visit(Expr.Nested node) {
			if (enter(node)) {
				node.expr.accept(this);
				leave(node);
			}
		}
		boolean enter(Expr.Add node) {
			return true;
		}
		void leave(Expr.Add node) {}
		void visit(Expr.Add node) {
			if (enter(node)) {
				node.leftTerm.accept(this);
				node.rightTerm.accept(this);
				leave(node);
			}
		}
		boolean enter(Expr.Sub node) {
			return true;
		}
		void leave(Expr.Sub node) {}
		void visit(Expr.Sub node) {
			if (enter(node)) {
				node.minuend.accept(this);
				node.subtrahend.accept(this);
				leave(node);
			}
		}
		boolean enter(Expr.Mul node) {
			return true;
		}
		void leave(Expr.Mul node) {}
		void visit(Expr.Mul node) {
			if (enter(node)) {
				node.leftFactor.accept(this);
				node.rightFactor.accept(this);
				leave(node);
			}
		}
		boolean enter(Expr.Div node) {
			return true;
		}
		void leave(Expr.Div node) {}
		void visit(Expr.Div node) {
			if (enter(node)) {
				node.dividend.accept(this);
				node.divisor.accept(this);
				leave(node);
			}
		}
	}
    // End of generated AST

	static class Calculator extends Walker {
	    static class Stack {
	        BigDecimal[] values = new BigDecimal[16];
	        int top = -1;
	        void push(BigDecimal num) {
	            if (num == null)
	                throw new IllegalArgumentException();
	            values[++top] = num;
	        }
	        BigDecimal pop() {
	            if (top < 0)
	                throw new IllegalStateException("empty");
	            return values[top--];
	        }
	    }
	    Map vars = new TreeMap();
	    Stack stack = new Stack();
	    String result;

	    void leave(Stmt.Print node) {
	        result = stack.pop().toString();
	    }
        void leave(Stmt.Set node) {
            vars.put(node.id.text, stack.pop());
        }
        void leave(Expr.Number node) {
            stack.push(new BigDecimal(node.num.text.toString()));
        }
        void leave(Expr.Var node) {
            stack.push((BigDecimal) vars.get(node.id.text));
        }
        void leave(Expr.Nested node) {
            // do nothing (otherwise we'd pop and immediately push back the popped value)
        }
        void leave(Expr.Add node) {
            BigDecimal rightArg = stack.pop();
            BigDecimal leftArg = stack.pop();
            stack.push(leftArg.add(rightArg));
        }
        void leave(Expr.Sub node) {
            BigDecimal rightArg = stack.pop();
            BigDecimal leftArg = stack.pop();
            stack.push(leftArg.subtract(rightArg));
        }
        void leave(Expr.Mul node) {
            BigDecimal rightArg = stack.pop();
            BigDecimal leftArg = stack.pop();
            stack.push(leftArg.multiply(rightArg));
        }
        void leave(Expr.Div node) {
            BigDecimal rightArg = stack.pop();
            BigDecimal leftArg = stack.pop();
            stack.push(leftArg.divide(rightArg));
        }
	}

    Grammar grammar;
    ParserState firstState;

    public JavaClassParserWriterTest(String name) {
        super(name);

        grammar = new GrammarTest(name).grammar;

//        PrintWriter out = new PrintWriter(System.out);
//        new JavaASTWriter(grammar).writeAST(out);
//        out.flush();

        firstState = new ParserState.Builder().buildParserStates(grammar);
        Collection resolvedConflictsLog = new ArrayList();
        ParserState.Builder.resolveConflicts(firstState, resolvedConflictsLog);
    }

    public void test() throws Exception {
        final byte[] bc = new JavaClassParserWriter().generate("beaver.cc.parser.ExprParser", "beaver.Parser", "beaver.cc.parser.JavaClassParserWriterTest", grammar, firstState);
        ClassLoader ldr = new ClassLoader() {
            public Class loadClass(String name) throws ClassNotFoundException {
                if (name.equals("beaver.cc.parser.ExprParser")) {
                    return defineClass(name, bc, 0, bc.length);
                } else {
                    return super.loadClass(name);
                }
            }
        };
        Class parserClass = ldr.loadClass("beaver.cc.parser.ExprParser");
        Constructor constr = parserClass.getConstructor(null);
        Parser parser = (Parser) constr.newInstance(null);
        final Map terms = new TreeMap();
        for (int i = 0; i < grammar.terminals.length; i++) {
            terms.put(grammar.terminals[i].getName(), grammar.terminals[i]);
        }
        Scanner scanner = new Scanner() {
            /*
             * t = (128.3 + 172.4) * 1.5 - 289.7775 / 12.15;
             * print 1.4 + t / (t - 27.2);
             */
            beaver.Term[]  tokens = {
                    new Term(((Symbol.Terminal) terms.get("ID")).id, 1, 1, "t"),
                    new Term(((Symbol.Terminal) terms.get("\"=\"")).id, 1, 3, null),
                    new Term(((Symbol.Terminal) terms.get("\"(\"")).id, 1, 5, null),
                    new Term(((Symbol.Terminal) terms.get("NUM")).id, 1, 6, "128.3"),
                    new Term(((Symbol.Terminal) terms.get("\"+\"")).id, 1, 12, null),
                    new Term(((Symbol.Terminal) terms.get("NUM")).id, 1, 14, "172.4"),
                    new Term(((Symbol.Terminal) terms.get("\")\"")).id, 1, 19, null),
                    new Term(((Symbol.Terminal) terms.get("\"*\"")).id, 1, 21, null),
                    new Term(((Symbol.Terminal) terms.get("NUM")).id, 1, 23, "1.5"),
                    new Term(((Symbol.Terminal) terms.get("\"-\"")).id, 1, 27, null),
                    new Term(((Symbol.Terminal) terms.get("NUM")).id, 1, 29, "289.7775"),
                    new Term(((Symbol.Terminal) terms.get("\"/\"")).id, 1, 35, null),
                    new Term(((Symbol.Terminal) terms.get("NUM")).id, 1, 37, "12.15"),
                    new Term(((Symbol.Terminal) terms.get("\";\"")).id, 1, 41, null),
                    new Term(((Symbol.Terminal) terms.get("\"print\"")).id, 2, 1, null),
                    new Term(((Symbol.Terminal) terms.get("NUM")).id, 2, 7, "1.4"),
                    new Term(((Symbol.Terminal) terms.get("\"+\"")).id, 2, 12, null),
                    new Term(((Symbol.Terminal) terms.get("ID")).id, 2, 14, "t"),
                    new Term(((Symbol.Terminal) terms.get("\"/\"")).id, 2, 16, null),
                    new Term(((Symbol.Terminal) terms.get("\"(\"")).id, 2, 18, null),
                    new Term(((Symbol.Terminal) terms.get("ID")).id, 2, 19, "t"),
                    new Term(((Symbol.Terminal) terms.get("\"-\"")).id, 2, 21, null),
                    new Term(((Symbol.Terminal) terms.get("NUM")).id, 2, 23, "27.2"),
                    new Term(((Symbol.Terminal) terms.get("\")\"")).id, 2, 27, null),
                    new Term(((Symbol.Terminal) terms.get("\";\"")).id, 2, 28, null),
            };
            int            cursor;
            beaver.Term    lastToken;

            public void setTokenInspector(TokenInspector insp) {
            }

            public beaver.Term makeTerm(int id, int line, int column, Object text) {
                return new Term(id, line, column, text);
            }

            public beaver.Term getNextToken() throws IOException {
                return cursor < tokens.length ? tokens[cursor++] : makeTerm(0, 3, 1, null);
            }

            public beaver.Term getLastToken() {
                return lastToken;
            }
        };
        Object root = parser.parse(scanner);
        assertTrue(root instanceof StmtList);
        Calculator eval = new Calculator();
        ((StmtList) root).accept(eval);
        assertNotNull(eval.result);
        assertEquals("2.468", eval.result);
    }

}
