/**
 * Beaver: a compiler front-end compiler
 * Copyright (c) 2003-2012, Alexander Demenchuk
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
package beaver.cc.parser;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Map;
import java.util.TreeMap;

import junit.framework.TestCase;

public class ParserStateBuilderTest extends TestCase {

    Grammar grammar;

    public ParserStateBuilderTest(String name) {
        super(name);
        grammar = new GrammarTest(name).grammar;
    }

    static Map getShiftActions(ParserState st) {
        Map actions = new TreeMap();
        for (ParserAction act = st.firstShift; act != null; act = act.next) {
            assertTrue(act instanceof ParserAction.Shift);
            ParserAction.Shift shift = (ParserAction.Shift) act;
            actions.put(shift.lookahead.toString(), shift.dest);
        }
        return actions;
    }

    static Map getReduceActions(ParserState st) {
        Map actions = new TreeMap();
        for (ParserAction act = st.firstReduce; act != null; act = act.next) {
            assertTrue(act instanceof ParserAction.Reduce);
            ParserAction.Reduce reduce = (ParserAction.Reduce) act;
            actions.put(reduce.lookahead.toString(), reduce.production);
        }
        return actions;
    }

    static ParserState[] toArray(ParserState st) {
        ParserState lastState = st;
        while (lastState.next != null) {
            lastState = lastState.next;
        }
        ParserState[] array = new ParserState[lastState.id + 1];
        while (st != null) {
            array[st.id] = st;
            st = st.next;
        }
        return array;
    }

    public void testInitialStates() {
        ParserState[] states = toArray(new ParserState.Builder().buildParserStates(grammar));
        assertEquals(24, states.length);

        Map shifts, reductions;

        ParserState st = states[0];
        assertEquals(0, st.id);
        assertNotNull(st.firstShift);
        shifts = getShiftActions(st);
        assertEquals(4, shifts.size());
        assertTrue(shifts.containsKey("ID"));
        assertEquals(19, ((ParserState) shifts.get("ID")).id);
        assertTrue(shifts.containsKey("\"print\""));
        assertEquals( 3, ((ParserState) shifts.get("\"print\"")).id);
        assertTrue(shifts.containsKey("Stmt"));
        assertEquals(23, ((ParserState) shifts.get("Stmt")).id);
        assertTrue(shifts.containsKey("StmtList"));
        assertEquals( 1, ((ParserState) shifts.get("StmtList")).id);
        assertNull(st.firstReduce);
        assertNull(st.defaultReduce);
        assertNotNull(st.accept);
        assertEquals("Eval", st.accept.lookahead.toString());

        st = states[1];
        assertEquals(1, st.id);
        assertNotNull(st.firstShift);
        shifts = getShiftActions(st);
        assertEquals(3, shifts.size());
        assertTrue(shifts.containsKey("ID"));
        assertEquals(19, ((ParserState) shifts.get("ID")).id);
        assertTrue(shifts.containsKey("\"print\""));
        assertEquals( 3, ((ParserState) shifts.get("\"print\"")).id);
        assertTrue(shifts.containsKey("Stmt"));
        assertEquals( 2, ((ParserState) shifts.get("Stmt")).id);
        assertNotNull(st.firstReduce);
        reductions = getReduceActions(st);
        assertEquals(1, reductions.size());
        assertTrue(reductions.containsKey("EOF"));
        assertEquals(grammar.productions[0].id, ((Production) reductions.get("EOF")).id);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[2];
        assertEquals(2, st.id);
        assertNull(st.firstShift);
        assertNotNull(st.firstReduce);
        reductions = getReduceActions(st);
        assertEquals(3, reductions.size());
        assertTrue(reductions.containsKey("ID"));
        assertEquals(grammar.productions[2].id, ((Production) reductions.get("ID")).id);
        assertTrue(reductions.containsKey("\"print\""));
        assertEquals(grammar.productions[2].id, ((Production) reductions.get("\"print\"")).id);
        assertTrue(reductions.containsKey("EOF"));
        assertEquals(grammar.productions[2].id, ((Production) reductions.get("EOF")).id);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[3];
        assertEquals(3, st.id);
        assertNotNull(st.firstShift);
        shifts = getShiftActions(st);
        assertEquals(4, shifts.size());
        assertTrue(shifts.containsKey("\"(\""));
        assertEquals(16, ((ParserState) shifts.get("\"(\"")).id);
        assertTrue(shifts.containsKey("ID"));
        assertEquals(15, ((ParserState) shifts.get("ID")).id);
        assertTrue(shifts.containsKey("NUM"));
        assertEquals(14, ((ParserState) shifts.get("NUM")).id);
        assertTrue(shifts.containsKey("Expr"));
        assertEquals( 4, ((ParserState) shifts.get("Expr")).id);
        assertNull(st.firstReduce);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[4];
        assertEquals(4, st.id);
        assertNotNull(st.firstShift);
        shifts = getShiftActions(st);
        assertEquals(5, shifts.size());
        assertTrue(shifts.containsKey("\"/\""));
        assertEquals(12, ((ParserState) shifts.get("\"/\"")).id);
        assertTrue(shifts.containsKey("\"*\""));
        assertEquals(10, ((ParserState) shifts.get("\"*\"")).id);
        assertTrue(shifts.containsKey("\"-\""));
        assertEquals( 8, ((ParserState) shifts.get("\"-\"")).id);
        assertTrue(shifts.containsKey("\"+\""));
        assertEquals( 6, ((ParserState) shifts.get("\"+\"")).id);
        assertTrue(shifts.containsKey("\";\""));
        assertEquals( 5, ((ParserState) shifts.get("\";\"")).id);
        assertNull(st.firstReduce);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[5];
        assertEquals(5, st.id);
        assertNull(st.firstShift);
        assertNotNull(st.firstReduce);
        reductions = getReduceActions(st);
        assertEquals(3, reductions.size());
        assertTrue(reductions.containsKey("ID"));
        assertEquals(grammar.productions[3].id, ((Production) reductions.get("ID")).id);
        assertTrue(reductions.containsKey("\"print\""));
        assertEquals(grammar.productions[3].id, ((Production) reductions.get("\"print\"")).id);
        assertTrue(reductions.containsKey("EOF"));
        assertEquals(grammar.productions[3].id, ((Production) reductions.get("EOF")).id);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[6];
        assertEquals(6, st.id);
        assertNotNull(st.firstShift);
        shifts = getShiftActions(st);
        assertEquals(4, shifts.size());
        assertTrue(shifts.containsKey("\"(\""));
        assertEquals(16, ((ParserState) shifts.get("\"(\"")).id);
        assertTrue(shifts.containsKey("ID"));
        assertEquals(15, ((ParserState) shifts.get("ID")).id);
        assertTrue(shifts.containsKey("NUM"));
        assertEquals(14, ((ParserState) shifts.get("NUM")).id);
        assertTrue(shifts.containsKey("Expr"));
        assertEquals( 7, ((ParserState) shifts.get("Expr")).id);
        assertNull(st.firstReduce);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[7];
        assertEquals(7, st.id);
        assertNotNull(st.firstShift);
        shifts = getShiftActions(st);
        assertEquals(4, shifts.size());
        assertTrue(shifts.containsKey("\"/\""));
        assertEquals(12, ((ParserState) shifts.get("\"/\"")).id);
        assertTrue(shifts.containsKey("\"*\""));
        assertEquals(10, ((ParserState) shifts.get("\"*\"")).id);
        assertTrue(shifts.containsKey("\"-\""));
        assertEquals( 8, ((ParserState) shifts.get("\"-\"")).id);
        assertTrue(shifts.containsKey("\"+\""));
        assertEquals( 6, ((ParserState) shifts.get("\"+\"")).id);
        assertNotNull(st.firstReduce);
        reductions = getReduceActions(st);
        assertEquals(6, reductions.size()); // including 4 conflicts
        assertTrue(reductions.containsKey("\"/\""));
        assertEquals(grammar.productions[8].id, ((Production) reductions.get("\"/\"")).id);
        assertTrue(reductions.containsKey("\"*\""));
        assertEquals(grammar.productions[8].id, ((Production) reductions.get("\"*\"")).id);
        assertTrue(reductions.containsKey("\"-\""));
        assertEquals(grammar.productions[8].id, ((Production) reductions.get("\"-\"")).id);
        assertTrue(reductions.containsKey("\"+\""));
        assertEquals(grammar.productions[8].id, ((Production) reductions.get("\"+\"")).id);
        assertTrue(reductions.containsKey("\")\""));
        assertEquals(grammar.productions[8].id, ((Production) reductions.get("\")\"")).id);
        assertTrue(reductions.containsKey("\";\""));
        assertEquals(grammar.productions[8].id, ((Production) reductions.get("\";\"")).id);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[8];
        assertEquals(8, st.id);
        assertNotNull(st.firstShift);
        shifts = getShiftActions(st);
        assertEquals(4, shifts.size());
        assertTrue(shifts.containsKey("\"(\""));
        assertEquals(16, ((ParserState) shifts.get("\"(\"")).id);
        assertTrue(shifts.containsKey("ID"));
        assertEquals(15, ((ParserState) shifts.get("ID")).id);
        assertTrue(shifts.containsKey("NUM"));
        assertEquals(14, ((ParserState) shifts.get("NUM")).id);
        assertTrue(shifts.containsKey("Expr"));
        assertEquals( 9, ((ParserState) shifts.get("Expr")).id);
        assertNull(st.firstReduce);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[9];
        assertEquals(9, st.id);
        assertNotNull(st.firstShift);
        shifts = getShiftActions(st);
        assertEquals(4, shifts.size());
        assertTrue(shifts.containsKey("\"/\""));
        assertEquals(12, ((ParserState) shifts.get("\"/\"")).id);
        assertTrue(shifts.containsKey("\"*\""));
        assertEquals(10, ((ParserState) shifts.get("\"*\"")).id);
        assertTrue(shifts.containsKey("\"-\""));
        assertEquals( 8, ((ParserState) shifts.get("\"-\"")).id);
        assertTrue(shifts.containsKey("\"+\""));
        assertEquals( 6, ((ParserState) shifts.get("\"+\"")).id);
        assertNotNull(st.firstReduce);
        reductions = getReduceActions(st);
        assertEquals(6, reductions.size()); // including 4 conflicts
        assertTrue(reductions.containsKey("\"/\""));
        assertEquals(grammar.productions[9].id, ((Production) reductions.get("\"/\"")).id);
        assertTrue(reductions.containsKey("\"*\""));
        assertEquals(grammar.productions[9].id, ((Production) reductions.get("\"*\"")).id);
        assertTrue(reductions.containsKey("\"-\""));
        assertEquals(grammar.productions[9].id, ((Production) reductions.get("\"-\"")).id);
        assertTrue(reductions.containsKey("\"+\""));
        assertEquals(grammar.productions[9].id, ((Production) reductions.get("\"+\"")).id);
        assertTrue(reductions.containsKey("\")\""));
        assertEquals(grammar.productions[9].id, ((Production) reductions.get("\")\"")).id);
        assertTrue(reductions.containsKey("\";\""));
        assertEquals(grammar.productions[9].id, ((Production) reductions.get("\";\"")).id);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[10];
        assertEquals(10, st.id);
        assertNotNull(st.firstShift);
        shifts = getShiftActions(st);
        assertEquals(4, shifts.size());
        assertTrue(shifts.containsKey("\"(\""));
        assertEquals(16, ((ParserState) shifts.get("\"(\"")).id);
        assertTrue(shifts.containsKey("ID"));
        assertEquals(15, ((ParserState) shifts.get("ID")).id);
        assertTrue(shifts.containsKey("NUM"));
        assertEquals(14, ((ParserState) shifts.get("NUM")).id);
        assertTrue(shifts.containsKey("Expr"));
        assertEquals(11, ((ParserState) shifts.get("Expr")).id);
        assertNull(st.firstReduce);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[11];
        assertEquals(11, st.id);
        assertNotNull(st.firstShift);
        shifts = getShiftActions(st);
        assertEquals(4, shifts.size());
        assertTrue(shifts.containsKey("\"/\""));
        assertEquals(12, ((ParserState) shifts.get("\"/\"")).id);
        assertTrue(shifts.containsKey("\"*\""));
        assertEquals(10, ((ParserState) shifts.get("\"*\"")).id);
        assertTrue(shifts.containsKey("\"-\""));
        assertEquals( 8, ((ParserState) shifts.get("\"-\"")).id);
        assertTrue(shifts.containsKey("\"+\""));
        assertEquals( 6, ((ParserState) shifts.get("\"+\"")).id);
        assertNotNull(st.firstReduce);
        reductions = getReduceActions(st);
        assertEquals(6, reductions.size()); // including 4 conflicts
        assertTrue(reductions.containsKey("\"/\""));
        assertEquals(grammar.productions[10].id, ((Production) reductions.get("\"/\"")).id);
        assertTrue(reductions.containsKey("\"*\""));
        assertEquals(grammar.productions[10].id, ((Production) reductions.get("\"*\"")).id);
        assertTrue(reductions.containsKey("\"-\""));
        assertEquals(grammar.productions[10].id, ((Production) reductions.get("\"-\"")).id);
        assertTrue(reductions.containsKey("\"+\""));
        assertEquals(grammar.productions[10].id, ((Production) reductions.get("\"+\"")).id);
        assertTrue(reductions.containsKey("\")\""));
        assertEquals(grammar.productions[10].id, ((Production) reductions.get("\")\"")).id);
        assertTrue(reductions.containsKey("\";\""));
        assertEquals(grammar.productions[10].id, ((Production) reductions.get("\";\"")).id);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[12];
        assertEquals(12, st.id);
        assertNotNull(st.firstShift);
        shifts = getShiftActions(st);
        assertEquals(4, shifts.size());
        assertTrue(shifts.containsKey("\"(\""));
        assertEquals(16, ((ParserState) shifts.get("\"(\"")).id);
        assertTrue(shifts.containsKey("ID"));
        assertEquals(15, ((ParserState) shifts.get("ID")).id);
        assertTrue(shifts.containsKey("NUM"));
        assertEquals(14, ((ParserState) shifts.get("NUM")).id);
        assertTrue(shifts.containsKey("Expr"));
        assertEquals(13, ((ParserState) shifts.get("Expr")).id);
        assertNull(st.firstReduce);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[13];
        assertEquals(13, st.id);
        assertNotNull(st.firstShift);
        shifts = getShiftActions(st);
        assertEquals(4, shifts.size());
        assertTrue(shifts.containsKey("\"/\""));
        assertEquals(12, ((ParserState) shifts.get("\"/\"")).id);
        assertTrue(shifts.containsKey("\"*\""));
        assertEquals(10, ((ParserState) shifts.get("\"*\"")).id);
        assertTrue(shifts.containsKey("\"-\""));
        assertEquals( 8, ((ParserState) shifts.get("\"-\"")).id);
        assertTrue(shifts.containsKey("\"+\""));
        assertEquals( 6, ((ParserState) shifts.get("\"+\"")).id);
        assertNotNull(st.firstReduce);
        reductions = getReduceActions(st);
        assertEquals(6, reductions.size()); // including 4 conflicts
        assertTrue(reductions.containsKey("\"/\""));
        assertEquals(grammar.productions[11].id, ((Production) reductions.get("\"/\"")).id);
        assertTrue(reductions.containsKey("\"*\""));
        assertEquals(grammar.productions[11].id, ((Production) reductions.get("\"*\"")).id);
        assertTrue(reductions.containsKey("\"-\""));
        assertEquals(grammar.productions[11].id, ((Production) reductions.get("\"-\"")).id);
        assertTrue(reductions.containsKey("\"+\""));
        assertEquals(grammar.productions[11].id, ((Production) reductions.get("\"+\"")).id);
        assertTrue(reductions.containsKey("\")\""));
        assertEquals(grammar.productions[11].id, ((Production) reductions.get("\")\"")).id);
        assertTrue(reductions.containsKey("\";\""));
        assertEquals(grammar.productions[11].id, ((Production) reductions.get("\";\"")).id);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[14];
        assertEquals(14, st.id);
        assertNull(st.firstShift);
        assertNotNull(st.firstReduce);
        reductions = getReduceActions(st);
        assertEquals(6, reductions.size());
        assertTrue(reductions.containsKey("\"/\""));
        assertEquals(grammar.productions[5].id, ((Production) reductions.get("\"/\"")).id);
        assertTrue(reductions.containsKey("\"*\""));
        assertEquals(grammar.productions[5].id, ((Production) reductions.get("\"*\"")).id);
        assertTrue(reductions.containsKey("\"-\""));
        assertEquals(grammar.productions[5].id, ((Production) reductions.get("\"-\"")).id);
        assertTrue(reductions.containsKey("\"+\""));
        assertEquals(grammar.productions[5].id, ((Production) reductions.get("\"+\"")).id);
        assertTrue(reductions.containsKey("\")\""));
        assertEquals(grammar.productions[5].id, ((Production) reductions.get("\")\"")).id);
        assertTrue(reductions.containsKey("\";\""));
        assertEquals(grammar.productions[5].id, ((Production) reductions.get("\";\"")).id);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[15];
        assertEquals(15, st.id);
        assertNull(st.firstShift);
        assertNotNull(st.firstReduce);
        reductions = getReduceActions(st);
        assertEquals(6, reductions.size());
        assertTrue(reductions.containsKey("\"/\""));
        assertEquals(grammar.productions[6].id, ((Production) reductions.get("\"/\"")).id);
        assertTrue(reductions.containsKey("\"*\""));
        assertEquals(grammar.productions[6].id, ((Production) reductions.get("\"*\"")).id);
        assertTrue(reductions.containsKey("\"-\""));
        assertEquals(grammar.productions[6].id, ((Production) reductions.get("\"-\"")).id);
        assertTrue(reductions.containsKey("\"+\""));
        assertEquals(grammar.productions[6].id, ((Production) reductions.get("\"+\"")).id);
        assertTrue(reductions.containsKey("\")\""));
        assertEquals(grammar.productions[6].id, ((Production) reductions.get("\")\"")).id);
        assertTrue(reductions.containsKey("\";\""));
        assertEquals(grammar.productions[6].id, ((Production) reductions.get("\";\"")).id);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[16];
        assertEquals(16, st.id);
        assertNotNull(st.firstShift);
        shifts = getShiftActions(st);
        assertEquals(4, shifts.size());
        assertTrue(shifts.containsKey("\"(\""));
        assertEquals(16, ((ParserState) shifts.get("\"(\"")).id);
        assertTrue(shifts.containsKey("ID"));
        assertEquals(15, ((ParserState) shifts.get("ID")).id);
        assertTrue(shifts.containsKey("NUM"));
        assertEquals(14, ((ParserState) shifts.get("NUM")).id);
        assertTrue(shifts.containsKey("Expr"));
        assertEquals(17, ((ParserState) shifts.get("Expr")).id);
        assertNull(st.firstReduce);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[17];
        assertEquals(17, st.id);
        assertNotNull(st.firstShift);
        shifts = getShiftActions(st);
        assertEquals(5, shifts.size());
        assertTrue(shifts.containsKey("\"/\""));
        assertEquals(12, ((ParserState) shifts.get("\"/\"")).id);
        assertTrue(shifts.containsKey("\"*\""));
        assertEquals(10, ((ParserState) shifts.get("\"*\"")).id);
        assertTrue(shifts.containsKey("\"-\""));
        assertEquals( 8, ((ParserState) shifts.get("\"-\"")).id);
        assertTrue(shifts.containsKey("\"+\""));
        assertEquals( 6, ((ParserState) shifts.get("\"+\"")).id);
        assertTrue(shifts.containsKey("\")\""));
        assertEquals(18, ((ParserState) shifts.get("\")\"")).id);
        assertNull(st.firstReduce);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[18];
        assertEquals(18, st.id);
        assertNull(st.firstShift);
        assertNotNull(st.firstReduce);
        reductions = getReduceActions(st);
        assertEquals(6, reductions.size());
        assertTrue(reductions.containsKey("\"/\""));
        assertEquals(grammar.productions[7].id, ((Production) reductions.get("\"/\"")).id);
        assertTrue(reductions.containsKey("\"*\""));
        assertEquals(grammar.productions[7].id, ((Production) reductions.get("\"*\"")).id);
        assertTrue(reductions.containsKey("\"-\""));
        assertEquals(grammar.productions[7].id, ((Production) reductions.get("\"-\"")).id);
        assertTrue(reductions.containsKey("\"+\""));
        assertEquals(grammar.productions[7].id, ((Production) reductions.get("\"+\"")).id);
        assertTrue(reductions.containsKey("\")\""));
        assertEquals(grammar.productions[7].id, ((Production) reductions.get("\")\"")).id);
        assertTrue(reductions.containsKey("\";\""));
        assertEquals(grammar.productions[7].id, ((Production) reductions.get("\";\"")).id);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[19];
        assertEquals(19, st.id);
        assertNotNull(st.firstShift);
        shifts = getShiftActions(st);
        assertEquals(1, shifts.size());
        assertTrue(shifts.containsKey("\"=\""));
        assertEquals(20, ((ParserState) shifts.get("\"=\"")).id);
        assertNull(st.firstReduce);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[20];
        assertEquals(20, st.id);
        assertNotNull(st.firstShift);
        shifts = getShiftActions(st);
        assertEquals(4, shifts.size());
        assertTrue(shifts.containsKey("\"(\""));
        assertEquals(16, ((ParserState) shifts.get("\"(\"")).id);
        assertTrue(shifts.containsKey("ID"));
        assertEquals(15, ((ParserState) shifts.get("ID")).id);
        assertTrue(shifts.containsKey("NUM"));
        assertEquals(14, ((ParserState) shifts.get("NUM")).id);
        assertTrue(shifts.containsKey("Expr"));
        assertEquals(21, ((ParserState) shifts.get("Expr")).id);
        assertNull(st.firstReduce);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[21];
        assertEquals(21, st.id);
        assertNotNull(st.firstShift);
        shifts = getShiftActions(st);
        assertEquals(5, shifts.size());
        assertTrue(shifts.containsKey("\"/\""));
        assertEquals(12, ((ParserState) shifts.get("\"/\"")).id);
        assertTrue(shifts.containsKey("\"*\""));
        assertEquals(10, ((ParserState) shifts.get("\"*\"")).id);
        assertTrue(shifts.containsKey("\"-\""));
        assertEquals( 8, ((ParserState) shifts.get("\"-\"")).id);
        assertTrue(shifts.containsKey("\"+\""));
        assertEquals( 6, ((ParserState) shifts.get("\"+\"")).id);
        assertTrue(shifts.containsKey("\";\""));
        assertEquals(22, ((ParserState) shifts.get("\";\"")).id);
        assertNull(st.firstReduce);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[22];
        assertEquals(22, st.id);
        assertNull(st.firstShift);
        assertNotNull(st.firstReduce);
        reductions = getReduceActions(st);
        assertEquals(3, reductions.size());
        assertTrue(reductions.containsKey("ID"));
        assertEquals(grammar.productions[4].id, ((Production) reductions.get("ID")).id);
        assertTrue(reductions.containsKey("\"print\""));
        assertEquals(grammar.productions[4].id, ((Production) reductions.get("\"print\"")).id);
        assertTrue(reductions.containsKey("EOF"));
        assertEquals(grammar.productions[4].id, ((Production) reductions.get("EOF")).id);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[23];
        assertEquals(23, st.id);
        assertNull(st.firstShift);
        assertNotNull(st.firstReduce);
        reductions = getReduceActions(st);
        assertEquals(3, reductions.size());
        assertTrue(reductions.containsKey("ID"));
        assertEquals(grammar.productions[1].id, ((Production) reductions.get("ID")).id);
        assertTrue(reductions.containsKey("\"print\""));
        assertEquals(grammar.productions[1].id, ((Production) reductions.get("\"print\"")).id);
        assertTrue(reductions.containsKey("EOF"));
        assertEquals(grammar.productions[1].id, ((Production) reductions.get("EOF")).id);
        assertNull(st.defaultReduce);
        assertNull(st.accept);
    }

    public void testConflictResolution() {
        ParserState[] states = toArray(new ParserState.Builder().buildParserStates(grammar));
        assertEquals(24, states.length);
        Collection resolvedConflictsLog = new ArrayList();
        ParserState.Builder.resolveConflicts(states[0], resolvedConflictsLog);
        assertEquals("Number of states where conflicts have been resolved", 4, resolvedConflictsLog.size());

        Map shifts, reductions;

        ParserState st = states[0];
        assertEquals(0, st.id);
        assertNotNull(st.firstShift);
        shifts = getShiftActions(st);
        assertEquals(4, shifts.size());
        assertTrue(shifts.containsKey("ID"));
        assertEquals(19, ((ParserState) shifts.get("ID")).id);
        assertTrue(shifts.containsKey("\"print\""));
        assertEquals( 3, ((ParserState) shifts.get("\"print\"")).id);
        assertTrue(shifts.containsKey("Stmt"));
        assertEquals(23, ((ParserState) shifts.get("Stmt")).id);
        assertTrue(shifts.containsKey("StmtList"));
        assertEquals( 1, ((ParserState) shifts.get("StmtList")).id);
        assertNull(st.firstReduce);
        assertNull(st.defaultReduce);
        assertNotNull(st.accept);
        assertEquals("Eval", st.accept.lookahead.toString());

        st = states[1];
        assertEquals(1, st.id);
        assertNotNull(st.firstShift);
        shifts = getShiftActions(st);
        assertEquals(3, shifts.size());
        assertTrue(shifts.containsKey("ID"));
        assertEquals(19, ((ParserState) shifts.get("ID")).id);
        assertTrue(shifts.containsKey("\"print\""));
        assertEquals( 3, ((ParserState) shifts.get("\"print\"")).id);
        assertTrue(shifts.containsKey("Stmt"));
        assertEquals( 2, ((ParserState) shifts.get("Stmt")).id);
        assertNotNull(st.firstReduce);
        reductions = getReduceActions(st);
        assertEquals(1, reductions.size());
        assertTrue(reductions.containsKey("EOF"));
        assertEquals(grammar.productions[0].id, ((Production) reductions.get("EOF")).id);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[2];
        assertEquals(2, st.id);
        assertNull(st.firstShift);
        assertNotNull(st.firstReduce);
        reductions = getReduceActions(st);
        assertEquals(3, reductions.size());
        assertTrue(reductions.containsKey("ID"));
        assertEquals(grammar.productions[2].id, ((Production) reductions.get("ID")).id);
        assertTrue(reductions.containsKey("\"print\""));
        assertEquals(grammar.productions[2].id, ((Production) reductions.get("\"print\"")).id);
        assertTrue(reductions.containsKey("EOF"));
        assertEquals(grammar.productions[2].id, ((Production) reductions.get("EOF")).id);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[3];
        assertEquals(3, st.id);
        assertNotNull(st.firstShift);
        shifts = getShiftActions(st);
        assertEquals(4, shifts.size());
        assertTrue(shifts.containsKey("\"(\""));
        assertEquals(16, ((ParserState) shifts.get("\"(\"")).id);
        assertTrue(shifts.containsKey("ID"));
        assertEquals(15, ((ParserState) shifts.get("ID")).id);
        assertTrue(shifts.containsKey("NUM"));
        assertEquals(14, ((ParserState) shifts.get("NUM")).id);
        assertTrue(shifts.containsKey("Expr"));
        assertEquals( 4, ((ParserState) shifts.get("Expr")).id);
        assertNull(st.firstReduce);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[4];
        assertEquals(4, st.id);
        assertNotNull(st.firstShift);
        shifts = getShiftActions(st);
        assertEquals(5, shifts.size());
        assertTrue(shifts.containsKey("\"/\""));
        assertEquals(12, ((ParserState) shifts.get("\"/\"")).id);
        assertTrue(shifts.containsKey("\"*\""));
        assertEquals(10, ((ParserState) shifts.get("\"*\"")).id);
        assertTrue(shifts.containsKey("\"-\""));
        assertEquals( 8, ((ParserState) shifts.get("\"-\"")).id);
        assertTrue(shifts.containsKey("\"+\""));
        assertEquals( 6, ((ParserState) shifts.get("\"+\"")).id);
        assertTrue(shifts.containsKey("\";\""));
        assertEquals( 5, ((ParserState) shifts.get("\";\"")).id);
        assertNull(st.firstReduce);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[5];
        assertEquals(5, st.id);
        assertNull(st.firstShift);
        assertNotNull(st.firstReduce);
        reductions = getReduceActions(st);
        assertEquals(3, reductions.size());
        assertTrue(reductions.containsKey("ID"));
        assertEquals(grammar.productions[3].id, ((Production) reductions.get("ID")).id);
        assertTrue(reductions.containsKey("\"print\""));
        assertEquals(grammar.productions[3].id, ((Production) reductions.get("\"print\"")).id);
        assertTrue(reductions.containsKey("EOF"));
        assertEquals(grammar.productions[3].id, ((Production) reductions.get("EOF")).id);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[6];
        assertEquals(6, st.id);
        assertNotNull(st.firstShift);
        shifts = getShiftActions(st);
        assertEquals(4, shifts.size());
        assertTrue(shifts.containsKey("\"(\""));
        assertEquals(16, ((ParserState) shifts.get("\"(\"")).id);
        assertTrue(shifts.containsKey("ID"));
        assertEquals(15, ((ParserState) shifts.get("ID")).id);
        assertTrue(shifts.containsKey("NUM"));
        assertEquals(14, ((ParserState) shifts.get("NUM")).id);
        assertTrue(shifts.containsKey("Expr"));
        assertEquals( 7, ((ParserState) shifts.get("Expr")).id);
        assertNull(st.firstReduce);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[7];
        assertEquals(7, st.id);
        assertNotNull(st.firstShift);
        shifts = getShiftActions(st);
        assertEquals(2, shifts.size());
        assertTrue(shifts.containsKey("\"/\""));
        assertEquals(12, ((ParserState) shifts.get("\"/\"")).id);
        assertTrue(shifts.containsKey("\"*\""));
        assertEquals(10, ((ParserState) shifts.get("\"*\"")).id);
        assertNotNull(st.firstReduce);
        reductions = getReduceActions(st);
        assertEquals(4, reductions.size());
        assertTrue(reductions.containsKey("\"-\""));
        assertEquals(grammar.productions[8].id, ((Production) reductions.get("\"-\"")).id);
        assertTrue(reductions.containsKey("\"+\""));
        assertEquals(grammar.productions[8].id, ((Production) reductions.get("\"+\"")).id);
        assertTrue(reductions.containsKey("\")\""));
        assertEquals(grammar.productions[8].id, ((Production) reductions.get("\")\"")).id);
        assertTrue(reductions.containsKey("\";\""));
        assertEquals(grammar.productions[8].id, ((Production) reductions.get("\";\"")).id);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[8];
        assertEquals(8, st.id);
        assertNotNull(st.firstShift);
        shifts = getShiftActions(st);
        assertEquals(4, shifts.size());
        assertTrue(shifts.containsKey("\"(\""));
        assertEquals(16, ((ParserState) shifts.get("\"(\"")).id);
        assertTrue(shifts.containsKey("ID"));
        assertEquals(15, ((ParserState) shifts.get("ID")).id);
        assertTrue(shifts.containsKey("NUM"));
        assertEquals(14, ((ParserState) shifts.get("NUM")).id);
        assertTrue(shifts.containsKey("Expr"));
        assertEquals( 9, ((ParserState) shifts.get("Expr")).id);
        assertNull(st.firstReduce);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[9];
        assertEquals(9, st.id);
        assertNotNull(st.firstShift);
        shifts = getShiftActions(st);
        assertEquals(2, shifts.size());
        assertTrue(shifts.containsKey("\"/\""));
        assertEquals(12, ((ParserState) shifts.get("\"/\"")).id);
        assertTrue(shifts.containsKey("\"*\""));
        assertEquals(10, ((ParserState) shifts.get("\"*\"")).id);
        assertNotNull(st.firstReduce);
        reductions = getReduceActions(st);
        assertEquals(4, reductions.size());
        assertTrue(reductions.containsKey("\"-\""));
        assertEquals(grammar.productions[9].id, ((Production) reductions.get("\"-\"")).id);
        assertTrue(reductions.containsKey("\"+\""));
        assertEquals(grammar.productions[9].id, ((Production) reductions.get("\"+\"")).id);
        assertTrue(reductions.containsKey("\")\""));
        assertEquals(grammar.productions[9].id, ((Production) reductions.get("\")\"")).id);
        assertTrue(reductions.containsKey("\";\""));
        assertEquals(grammar.productions[9].id, ((Production) reductions.get("\";\"")).id);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[10];
        assertEquals(10, st.id);
        assertNotNull(st.firstShift);
        shifts = getShiftActions(st);
        assertEquals(4, shifts.size());
        assertTrue(shifts.containsKey("\"(\""));
        assertEquals(16, ((ParserState) shifts.get("\"(\"")).id);
        assertTrue(shifts.containsKey("ID"));
        assertEquals(15, ((ParserState) shifts.get("ID")).id);
        assertTrue(shifts.containsKey("NUM"));
        assertEquals(14, ((ParserState) shifts.get("NUM")).id);
        assertTrue(shifts.containsKey("Expr"));
        assertEquals(11, ((ParserState) shifts.get("Expr")).id);
        assertNull(st.firstReduce);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[11];
        assertEquals(11, st.id);
        assertNull(st.firstShift);
        assertNotNull(st.firstReduce);
        reductions = getReduceActions(st);
        assertEquals(6, reductions.size());
        assertTrue(reductions.containsKey("\"/\""));
        assertEquals(grammar.productions[10].id, ((Production) reductions.get("\"/\"")).id);
        assertTrue(reductions.containsKey("\"*\""));
        assertEquals(grammar.productions[10].id, ((Production) reductions.get("\"*\"")).id);
        assertTrue(reductions.containsKey("\"-\""));
        assertEquals(grammar.productions[10].id, ((Production) reductions.get("\"-\"")).id);
        assertTrue(reductions.containsKey("\"+\""));
        assertEquals(grammar.productions[10].id, ((Production) reductions.get("\"+\"")).id);
        assertTrue(reductions.containsKey("\")\""));
        assertEquals(grammar.productions[10].id, ((Production) reductions.get("\")\"")).id);
        assertTrue(reductions.containsKey("\";\""));
        assertEquals(grammar.productions[10].id, ((Production) reductions.get("\";\"")).id);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[12];
        assertEquals(12, st.id);
        assertNotNull(st.firstShift);
        shifts = getShiftActions(st);
        assertEquals(4, shifts.size());
        assertTrue(shifts.containsKey("\"(\""));
        assertEquals(16, ((ParserState) shifts.get("\"(\"")).id);
        assertTrue(shifts.containsKey("ID"));
        assertEquals(15, ((ParserState) shifts.get("ID")).id);
        assertTrue(shifts.containsKey("NUM"));
        assertEquals(14, ((ParserState) shifts.get("NUM")).id);
        assertTrue(shifts.containsKey("Expr"));
        assertEquals(13, ((ParserState) shifts.get("Expr")).id);
        assertNull(st.firstReduce);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[13];
        assertEquals(13, st.id);
        assertNull(st.firstShift);
        assertNotNull(st.firstReduce);
        reductions = getReduceActions(st);
        assertEquals(6, reductions.size()); // including 4 conflicts
        assertTrue(reductions.containsKey("\"/\""));
        assertEquals(grammar.productions[11].id, ((Production) reductions.get("\"/\"")).id);
        assertTrue(reductions.containsKey("\"*\""));
        assertEquals(grammar.productions[11].id, ((Production) reductions.get("\"*\"")).id);
        assertTrue(reductions.containsKey("\"-\""));
        assertEquals(grammar.productions[11].id, ((Production) reductions.get("\"-\"")).id);
        assertTrue(reductions.containsKey("\"+\""));
        assertEquals(grammar.productions[11].id, ((Production) reductions.get("\"+\"")).id);
        assertTrue(reductions.containsKey("\")\""));
        assertEquals(grammar.productions[11].id, ((Production) reductions.get("\")\"")).id);
        assertTrue(reductions.containsKey("\";\""));
        assertEquals(grammar.productions[11].id, ((Production) reductions.get("\";\"")).id);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[14];
        assertEquals(14, st.id);
        assertNull(st.firstShift);
        assertNotNull(st.firstReduce);
        reductions = getReduceActions(st);
        assertEquals(6, reductions.size());
        assertTrue(reductions.containsKey("\"/\""));
        assertEquals(grammar.productions[5].id, ((Production) reductions.get("\"/\"")).id);
        assertTrue(reductions.containsKey("\"*\""));
        assertEquals(grammar.productions[5].id, ((Production) reductions.get("\"*\"")).id);
        assertTrue(reductions.containsKey("\"-\""));
        assertEquals(grammar.productions[5].id, ((Production) reductions.get("\"-\"")).id);
        assertTrue(reductions.containsKey("\"+\""));
        assertEquals(grammar.productions[5].id, ((Production) reductions.get("\"+\"")).id);
        assertTrue(reductions.containsKey("\")\""));
        assertEquals(grammar.productions[5].id, ((Production) reductions.get("\")\"")).id);
        assertTrue(reductions.containsKey("\";\""));
        assertEquals(grammar.productions[5].id, ((Production) reductions.get("\";\"")).id);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[15];
        assertEquals(15, st.id);
        assertNull(st.firstShift);
        assertNotNull(st.firstReduce);
        reductions = getReduceActions(st);
        assertEquals(6, reductions.size());
        assertTrue(reductions.containsKey("\"/\""));
        assertEquals(grammar.productions[6].id, ((Production) reductions.get("\"/\"")).id);
        assertTrue(reductions.containsKey("\"*\""));
        assertEquals(grammar.productions[6].id, ((Production) reductions.get("\"*\"")).id);
        assertTrue(reductions.containsKey("\"-\""));
        assertEquals(grammar.productions[6].id, ((Production) reductions.get("\"-\"")).id);
        assertTrue(reductions.containsKey("\"+\""));
        assertEquals(grammar.productions[6].id, ((Production) reductions.get("\"+\"")).id);
        assertTrue(reductions.containsKey("\")\""));
        assertEquals(grammar.productions[6].id, ((Production) reductions.get("\")\"")).id);
        assertTrue(reductions.containsKey("\";\""));
        assertEquals(grammar.productions[6].id, ((Production) reductions.get("\";\"")).id);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[16];
        assertEquals(16, st.id);
        assertNotNull(st.firstShift);
        shifts = getShiftActions(st);
        assertEquals(4, shifts.size());
        assertTrue(shifts.containsKey("\"(\""));
        assertEquals(16, ((ParserState) shifts.get("\"(\"")).id);
        assertTrue(shifts.containsKey("ID"));
        assertEquals(15, ((ParserState) shifts.get("ID")).id);
        assertTrue(shifts.containsKey("NUM"));
        assertEquals(14, ((ParserState) shifts.get("NUM")).id);
        assertTrue(shifts.containsKey("Expr"));
        assertEquals(17, ((ParserState) shifts.get("Expr")).id);
        assertNull(st.firstReduce);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[17];
        assertEquals(17, st.id);
        assertNotNull(st.firstShift);
        shifts = getShiftActions(st);
        assertEquals(5, shifts.size());
        assertTrue(shifts.containsKey("\"/\""));
        assertEquals(12, ((ParserState) shifts.get("\"/\"")).id);
        assertTrue(shifts.containsKey("\"*\""));
        assertEquals(10, ((ParserState) shifts.get("\"*\"")).id);
        assertTrue(shifts.containsKey("\"-\""));
        assertEquals( 8, ((ParserState) shifts.get("\"-\"")).id);
        assertTrue(shifts.containsKey("\"+\""));
        assertEquals( 6, ((ParserState) shifts.get("\"+\"")).id);
        assertTrue(shifts.containsKey("\")\""));
        assertEquals(18, ((ParserState) shifts.get("\")\"")).id);
        assertNull(st.firstReduce);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[18];
        assertEquals(18, st.id);
        assertNull(st.firstShift);
        assertNotNull(st.firstReduce);
        reductions = getReduceActions(st);
        assertEquals(6, reductions.size());
        assertTrue(reductions.containsKey("\"/\""));
        assertEquals(grammar.productions[7].id, ((Production) reductions.get("\"/\"")).id);
        assertTrue(reductions.containsKey("\"*\""));
        assertEquals(grammar.productions[7].id, ((Production) reductions.get("\"*\"")).id);
        assertTrue(reductions.containsKey("\"-\""));
        assertEquals(grammar.productions[7].id, ((Production) reductions.get("\"-\"")).id);
        assertTrue(reductions.containsKey("\"+\""));
        assertEquals(grammar.productions[7].id, ((Production) reductions.get("\"+\"")).id);
        assertTrue(reductions.containsKey("\")\""));
        assertEquals(grammar.productions[7].id, ((Production) reductions.get("\")\"")).id);
        assertTrue(reductions.containsKey("\";\""));
        assertEquals(grammar.productions[7].id, ((Production) reductions.get("\";\"")).id);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[19];
        assertEquals(19, st.id);
        assertNotNull(st.firstShift);
        shifts = getShiftActions(st);
        assertEquals(1, shifts.size());
        assertTrue(shifts.containsKey("\"=\""));
        assertEquals(20, ((ParserState) shifts.get("\"=\"")).id);
        assertNull(st.firstReduce);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[20];
        assertEquals(20, st.id);
        assertNotNull(st.firstShift);
        shifts = getShiftActions(st);
        assertEquals(4, shifts.size());
        assertTrue(shifts.containsKey("\"(\""));
        assertEquals(16, ((ParserState) shifts.get("\"(\"")).id);
        assertTrue(shifts.containsKey("ID"));
        assertEquals(15, ((ParserState) shifts.get("ID")).id);
        assertTrue(shifts.containsKey("NUM"));
        assertEquals(14, ((ParserState) shifts.get("NUM")).id);
        assertTrue(shifts.containsKey("Expr"));
        assertEquals(21, ((ParserState) shifts.get("Expr")).id);
        assertNull(st.firstReduce);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[21];
        assertEquals(21, st.id);
        assertNotNull(st.firstShift);
        shifts = getShiftActions(st);
        assertEquals(5, shifts.size());
        assertTrue(shifts.containsKey("\"/\""));
        assertEquals(12, ((ParserState) shifts.get("\"/\"")).id);
        assertTrue(shifts.containsKey("\"*\""));
        assertEquals(10, ((ParserState) shifts.get("\"*\"")).id);
        assertTrue(shifts.containsKey("\"-\""));
        assertEquals( 8, ((ParserState) shifts.get("\"-\"")).id);
        assertTrue(shifts.containsKey("\"+\""));
        assertEquals( 6, ((ParserState) shifts.get("\"+\"")).id);
        assertTrue(shifts.containsKey("\";\""));
        assertEquals(22, ((ParserState) shifts.get("\";\"")).id);
        assertNull(st.firstReduce);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[22];
        assertEquals(22, st.id);
        assertNull(st.firstShift);
        assertNotNull(st.firstReduce);
        reductions = getReduceActions(st);
        assertEquals(3, reductions.size());
        assertTrue(reductions.containsKey("ID"));
        assertEquals(grammar.productions[4].id, ((Production) reductions.get("ID")).id);
        assertTrue(reductions.containsKey("\"print\""));
        assertEquals(grammar.productions[4].id, ((Production) reductions.get("\"print\"")).id);
        assertTrue(reductions.containsKey("EOF"));
        assertEquals(grammar.productions[4].id, ((Production) reductions.get("EOF")).id);
        assertNull(st.defaultReduce);
        assertNull(st.accept);

        st = states[23];
        assertEquals(23, st.id);
        assertNull(st.firstShift);
        assertNotNull(st.firstReduce);
        reductions = getReduceActions(st);
        assertEquals(3, reductions.size());
        assertTrue(reductions.containsKey("ID"));
        assertEquals(grammar.productions[1].id, ((Production) reductions.get("ID")).id);
        assertTrue(reductions.containsKey("\"print\""));
        assertEquals(grammar.productions[1].id, ((Production) reductions.get("\"print\"")).id);
        assertTrue(reductions.containsKey("EOF"));
        assertEquals(grammar.productions[1].id, ((Production) reductions.get("EOF")).id);
        assertNull(st.defaultReduce);
        assertNull(st.accept);
    }
}
