/**
 * Beaver: a compiler front-end compiler
 * Copyright (c) 2003-2012, Alexander Demenchuk
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
package beaver.util;

import junit.framework.TestCase;

public class BitSetTest extends TestCase {

    public void testAddBitAtEdge() {
        BitSet set = new BitSet();
        int n = set.capacity();

        set.add(0);
        assertTrue(set.isSet(0));
        for (int i = 1; i < n; i++) {
            assertFalse(set.isSet(i));
        }
        assertEquals(1, set.size());

        n--;
        set.add(n);
        assertTrue(set.isSet(0));
        assertTrue(set.isSet(n));
        for (int i = 1; i < n; i++) {
            assertFalse(set.isSet(i));
        }
        assertEquals(2, set.size());
    }

    public void testAddBitWithinBoundaries() {
        BitSet set = new BitSet();

        set.add(25);
        assertEquals(1, set.size());
        assertTrue(set.isSet(25));

        set.add(80);
        assertEquals(2, set.size());
        assertTrue(set.isSet(80));

        set.add(50);
        assertEquals(3, set.size());
        assertTrue(set.isSet(50));
    }

    public void testAddBitBeyondBoundaries() {
        BitSet set = new BitSet(64);

        set.add(25);
        assertEquals(1, set.size());
        assertTrue(set.isSet(25));

        set.add(80);
        assertEquals(2, set.size());
        assertTrue(set.isSet(80));

        set.add(10);
        assertEquals(3, set.size());
        assertTrue(set.isSet(10));
        assertTrue(set.isSet(25));
        assertTrue(set.isSet(80));

        set.add(300);
        assertEquals(4, set.size());
        assertTrue(set.isSet(10));
        assertTrue(set.isSet(25));
        assertTrue(set.isSet(80));
        assertTrue(set.isSet(300));
    }

    public void testAddRangeToEmptySet() {
        BitSet set = new BitSet();

        set.add(12, 24);
        assertEquals(12, set.size());

        for (int i = 0; i < 12; i++) {
            assertFalse(set.isSet(i));
        }
        for (int i = 12; i < 24; i++) {
            assertTrue(set.isSet(i));
        }
        for (int i = 24; i < 256; i++) {
            assertFalse(set.isSet(i));
        }
    }

    public void testAddRangeToNonEmptySet() {
        BitSet set = new BitSet();

        set.add(10);
        set.add(15);
        set.add(20);
        set.add(25);
        assertEquals(4, set.size());

        set.add(12, 24);
        assertEquals(14, set.size());

        for (int i = 0; i < 10; i++) {
            assertFalse(set.isSet(i));
        }
        assertTrue(set.isSet(10));
        assertFalse(set.isSet(11));

        for (int i = 12; i < 24; i++) {
            assertTrue(set.isSet(i));
        }
        assertFalse(set.isSet(24));
        assertTrue(set.isSet(25));
        for (int i = 26; i < 256; i++) {
            assertFalse(set.isSet(i));
        }
    }

    public void testErase() {
        BitSet set = new BitSet();

        set.add(12, 24);
        set.erase(15);
        set.erase(20);
        assertEquals(10, set.size());

        for (int i = 12; i < 15; i++) {
            assertTrue(set.isSet(i));
        }
        assertFalse(set.isSet(15));
        for (int i = 16; i < 20; i++) {
            assertTrue(set.isSet(i));
        }
        assertFalse(set.isSet(20));
        for (int i = 21; i < 24; i++) {
            assertTrue(set.isSet(i));
        }
    }

    public void testClear() {
        BitSet set = new BitSet();

        set.add(0);
        set.add(10);
        set.add(12, 24);
        set.add(25);
        assertEquals(15, set.size());

        set.clear();
        assertTrue(set.isEmpty());
        assertEquals(0, set.size());

        int n = set.capacity();
        for (int i = 0; i < n; i++) {
            assertFalse(set.isSet(i));
        }
    }

    public void testAddBitSetToEmptySet() {
        BitSet srcSet = new BitSet();

        srcSet.add(10);
        srcSet.add(12, 24);
        srcSet.add(25);

        BitSet dstSet = new BitSet();
        assertTrue(dstSet.isEmpty());
        dstSet.add(srcSet);
        assertFalse(dstSet.isEmpty());
        assertEquals(14, dstSet.size());

        for (int i = 0; i < 10; i++) {
            assertFalse(dstSet.isSet(i));
        }
        assertTrue(dstSet.isSet(10));
        assertFalse(dstSet.isSet(11));
        for (int i = 12; i < 24; i++) {
            assertTrue(dstSet.isSet(i));
        }
        assertFalse(dstSet.isSet(24));
        assertTrue(dstSet.isSet(25));

        int n = dstSet.capacity();
        for (int i = 26; i < n; i++) {
            assertFalse(dstSet.isSet(i));
        }
    }

    public void testAddBitSetToFullSet() {
        BitSet srcSet = new BitSet();

        srcSet.add(10);
        srcSet.add(12, 24);
        srcSet.add(25);

        BitSet dstSet = new BitSet();

        srcSet.add(5);
        srcSet.add(15);
        srcSet.add(20);

        dstSet.add(srcSet);
        assertEquals(15, dstSet.size());

        for (int i = 0; i < 5; i++) {
            assertFalse(dstSet.isSet(i));
        }
        assertTrue(dstSet.isSet(5));
        for (int i = 6; i < 10; i++) {
            assertFalse(dstSet.isSet(i));
        }
        assertTrue(dstSet.isSet(10));
        assertFalse(dstSet.isSet(11));
        for (int i = 12; i < 24; i++) {
            assertTrue(dstSet.isSet(i));
        }
        assertFalse(dstSet.isSet(24));
        assertTrue(dstSet.isSet(25));

        int n = dstSet.capacity();
        for (int i = 26; i < n; i++) {
            assertFalse(dstSet.isSet(i));
        }
    }

    public void testAddBitSetToFullSetWithDifferentBoundaries() {
        BitSet srcSet = new BitSet(32);

        srcSet.add(10);
        srcSet.add(12, 24);
        srcSet.add(25);

        BitSet dstSet = new BitSet(32);

        srcSet.add(50);
        srcSet.add(64, 72);
        srcSet.add(80);

        dstSet.add(srcSet);
        assertEquals(24, dstSet.size());

        for (int i = 0; i < 10; i++) {
            assertFalse(dstSet.isSet(i));
        }
        assertTrue(dstSet.isSet(10));
        assertFalse(dstSet.isSet(11));
        for (int i = 12; i < 24; i++) {
            assertTrue(dstSet.isSet(i));
        }
        assertFalse(dstSet.isSet(24));
        assertTrue(dstSet.isSet(25));
        for (int i = 26; i < 50; i++) {
            assertFalse(dstSet.isSet(i));
        }
        assertTrue(dstSet.isSet(50));
        for (int i = 51; i < 64; i++) {
            assertFalse(dstSet.isSet(i));
        }
        for (int i = 64; i < 72; i++) {
            assertTrue(dstSet.isSet(i));
        }
        for (int i = 72; i < 80; i++) {
            assertFalse(dstSet.isSet(i));
        }
        assertTrue(dstSet.isSet(80));
        for (int i = 81; i < 100; i++) {
            assertFalse(dstSet.isSet(i));
        }
    }

    public void testIsSet() {
        BitSet set = new BitSet();
        set.add(5);
        assertTrue(set.isSet(5));
    }

    public void testIsEmpty() {
        BitSet set = new BitSet();
        assertTrue(set.isEmpty());
        set.add(10);
        assertFalse(set.isEmpty());
    }

    public void testSize() {
        BitSet set = new BitSet();
        assertEquals(0, set.size());
        set.add(10);
        assertEquals(1, set.size());
        set.add(20);
        assertEquals(2, set.size());
        set.add(15, 25);
        assertEquals(11, set.size());
        set.clear();
        assertEquals(0, set.size());
    }

    public void testCapacity() {
        BitSet set = new BitSet(32);
        assertEquals(32, set.capacity());
        set.add(0);
        set.add(32);
        assertTrue(set.capacity() > 32);
    }

    public void testForEachBitAccept() {
        BitSet set = new BitSet();
        set.add(5);
        set.add(10);
        set.add(15);
        final BitSet res = new BitSet();
        set.forEachBitAccept(new BitSet.BitVisitor() {
            public void visit(int bit) {
                res.add(bit);
            }
        });
        assertEquals(3, res.size());
        assertTrue(set.isSet(5));
        assertTrue(set.isSet(10));
        assertTrue(set.isSet(15));
    }

    public void testCountBits() {
        assertEquals(11, BitMath.countBits(0x7FC3));
        assertEquals(1, BitMath.countBits(0x8000));
    }
}
