package beaver.cc;
public abstract class AST {
	public static abstract class Node {
		Node parent;
		Node prev;
		Node next;
		abstract void accept(Walker walker);
		abstract Node dup();
		static String asString(Object object) {
			return object == null ? "" : " " + object.toString();
		}
	}
	public static class Term extends Node implements beaver.Term {
		static final String[] keywords = { "'<EOF>'", "'%grammar'", "'%tokens'", "'='", "';'", "'|'", "'{'", "'}'", "'('", "')'", "'%prec'", "':'", "','", "'%macros'", "'\\'", "'@'", "'>>'" };
		int    id;
		int    line;
		int    column;
		Object text;
		public Term(int id, int line, int column, Object text) {
			this.id = id;
			this.line = line;
			this.column = column;
			this.text = text;
		}
		private Term(Term term) {
			this.id = term.id;
			this.line = term.line;
			this.column = term.column;
			this.text = term.text;
		}
		public void setId(int newId) {
			id = newId;
		}
		public int getId() {
			return id;
		}
		public Object getText() {
			return text;
		}
		public int getLine() {
			return line;
		}
		public int getColumn() {
			return column;
		}
		void accept(Walker walker) {
			walker.visit(this);
		}
		Node dup() {
			return new Term(this);
		}
		public String toString() {
			return (0 <= id && id < keywords.length ? keywords[id] : text.toString());
		}
	}
	public static abstract class NodeList extends Node {
		Node first;
		Node last;
		int  size;
		NodeList() {}
		NodeList(Node node) {
			node.parent = this;
			last = first = node;
			size = 1;
		}
		protected NodeList(NodeList list) {
			Node elem = list.first;
			if (elem != null) {
				last = first = elem.dup();
				last.parent = this;
				last.prev = last.next = null;
				size = 1;
				for (elem = elem.next; elem != null; elem = elem.next) {
					add(elem.dup());
				}
			}
		}
		public void add(Node node) {
			node.parent = this;
			node.next = null;
			if ((node.prev = last) == null) {
				last = first = node;
			} else {
				last = last.next = node;
			}
			size++;
		}
		public void remove(Node node) {
			if (node.parent != this) {
				throw new IllegalArgumentException("wrong list");
			}
			if (node.prev != null) {
				first = node.next;
			} else {
				node.prev.next = node.next;
			}
			if (node.next == null) {
				last = node.prev;
			} else {
				node.next.prev = node.prev;
			}
			size--;
		}
		public void replace(Node node, Node newNode) {
			if (node.parent != this) {
				throw new IllegalArgumentException("wrong list");
			}
			newNode.parent = this;
			newNode.prev = node.prev;
			newNode.next = node.next;
			if (newNode.prev == null) {
				first = newNode;
			} else {
				newNode.prev.next = newNode;
			}
			if (newNode.next == null) {
				last = newNode;
			} else {
				newNode.next.prev = newNode;
			}
		}
		public String toString() {
			if (first != null) {
				String str = first.toString();
				for (Node e = first.next; e != null; e = e.next) {
					str += ", " + e;
				}
				return str;
			}
			return "";
		}
	}
	public static class Grammar extends Node {
		ProductionList productionList;
		PrecedenceList precedenceList;
		MacroList macroList;
		TokenList tokenList;
		public Grammar(ProductionList productionList, PrecedenceList precedenceList, MacroList macroList, TokenList tokenList) {
			(this.productionList = productionList).parent = this;
			(this.precedenceList = precedenceList).parent = this;
			(this.macroList = macroList).parent = this;
			(this.tokenList = tokenList).parent = this;
		}
		private Grammar(Grammar node) {
			(this.productionList = (ProductionList) node.productionList.dup()).parent = this;
			(this.precedenceList = (PrecedenceList) node.precedenceList.dup()).parent = this;
			(this.macroList = (MacroList) node.macroList.dup()).parent = this;
			(this.tokenList = (TokenList) node.tokenList.dup()).parent = this;
		}
		Node dup() {
			return new Grammar(this);
		}
		void accept(Walker walker) {
			walker.visit(this);
		}
		public String toString() {
			return "(Grammar" + " " + productionList + " " + precedenceList + " " + macroList + " " + tokenList + ")";
		}
	}
	public static class ProductionList extends NodeList {
		private ProductionList(ProductionList list) {
			super(list);
		}
		public ProductionList() {
			super();
		}
		public ProductionList(Production elem) {
			super(elem);
		}
		Node dup() {
			return new ProductionList(this);
		}
		void accept(Walker walker) {
			walker.visit(this);
		}
		public String toString() {
			return "(ProductionList " + super.toString() + ")";
		}
	}
	public static class Production extends Node {
		Term name;
		SymbolDefinitionList symbolDefinitionList;
		public Production(Term name, SymbolDefinitionList symbolDefinitionList) {
			(this.name = name).parent = this;
			(this.symbolDefinitionList = symbolDefinitionList).parent = this;
		}
		private Production(Production node) {
			(this.name = (Term) node.name.dup()).parent = this;
			(this.symbolDefinitionList = (SymbolDefinitionList) node.symbolDefinitionList.dup()).parent = this;
		}
		Node dup() {
			return new Production(this);
		}
		void accept(Walker walker) {
			walker.visit(this);
		}
		public String toString() {
			return "(Production" + " " + name + " " + symbolDefinitionList + ")";
		}
	}
	public static class SymbolDefinitionList extends NodeList {
		private SymbolDefinitionList(SymbolDefinitionList list) {
			super(list);
		}
		public SymbolDefinitionList() {
			super();
		}
		public SymbolDefinitionList(SymbolDefinition elem) {
			super(elem);
		}
		Node dup() {
			return new SymbolDefinitionList(this);
		}
		void accept(Walker walker) {
			walker.visit(this);
		}
		public String toString() {
			return "(SymbolDefinitionList " + super.toString() + ")";
		}
	}
	public static class SymbolDefinition extends Node {
		Term optRuleName;
		SymbolList symbolList;
		public SymbolDefinition(Term optRuleName, SymbolList symbolList) {
			if(optRuleName != null) {
				(this.optRuleName = optRuleName).parent = this;
			}
			(this.symbolList = symbolList).parent = this;
		}
		private SymbolDefinition(SymbolDefinition node) {
			if(node.optRuleName != null) {
				(this.optRuleName = (Term) node.optRuleName.dup()).parent = this;
			}
			(this.symbolList = (SymbolList) node.symbolList.dup()).parent = this;
		}
		Node dup() {
			return new SymbolDefinition(this);
		}
		void accept(Walker walker) {
			walker.visit(this);
		}
		public String toString() {
			return "(SymbolDefinition" + asString(optRuleName) + " " + symbolList + ")";
		}
	}
	public static class SymbolList extends NodeList {
		private SymbolList(SymbolList list) {
			super(list);
		}
		public SymbolList() {
			super();
		}
		public SymbolList(Symbol elem) {
			super(elem);
		}
		Node dup() {
			return new SymbolList(this);
		}
		void accept(Walker walker) {
			walker.visit(this);
		}
		public String toString() {
			return "(SymbolList " + super.toString() + ")";
		}
	}
	public static abstract class Symbol extends Node {
		public static class Name extends Symbol {
			Term name;
			Term optQuant;
			public Name(Term name, Term optQuant) {
				(this.name = name).parent = this;
				if(optQuant != null) {
					(this.optQuant = optQuant).parent = this;
				}
			}
			private Name(Name node) {
				(this.name = (Term) node.name.dup()).parent = this;
				if(node.optQuant != null) {
					(this.optQuant = (Term) node.optQuant.dup()).parent = this;
				}
			}
			Node dup() {
				return new Name(this);
			}
			void accept(Walker walker) {
				walker.visit(this);
			}
			public String toString() {
				return "(Symbol.Name" + " " + name + asString(optQuant) + ")";
			}
		}
		public static class Text extends Symbol {
			Term text;
			Term optQuant;
			public Text(Term text, Term optQuant) {
				(this.text = text).parent = this;
				if(optQuant != null) {
					(this.optQuant = optQuant).parent = this;
				}
			}
			private Text(Text node) {
				(this.text = (Term) node.text.dup()).parent = this;
				if(node.optQuant != null) {
					(this.optQuant = (Term) node.optQuant.dup()).parent = this;
				}
			}
			Node dup() {
				return new Text(this);
			}
			void accept(Walker walker) {
				walker.visit(this);
			}
			public String toString() {
				return "(Symbol.Text" + " " + text + asString(optQuant) + ")";
			}
		}
		public static class Nested extends Symbol {
			SymbolList symbolList;
			Term optQuant;
			public Nested(SymbolList symbolList, Term optQuant) {
				(this.symbolList = symbolList).parent = this;
				if(optQuant != null) {
					(this.optQuant = optQuant).parent = this;
				}
			}
			private Nested(Nested node) {
				(this.symbolList = (SymbolList) node.symbolList.dup()).parent = this;
				if(node.optQuant != null) {
					(this.optQuant = (Term) node.optQuant.dup()).parent = this;
				}
			}
			Node dup() {
				return new Nested(this);
			}
			void accept(Walker walker) {
				walker.visit(this);
			}
			public String toString() {
				return "(Symbol.Nested" + " " + symbolList + asString(optQuant) + ")";
			}
		}
	}
	public static class PrecedenceList extends NodeList {
		private PrecedenceList(PrecedenceList list) {
			super(list);
		}
		public PrecedenceList() {
			super();
		}
		public PrecedenceList(Precedence elem) {
			super(elem);
		}
		Node dup() {
			return new PrecedenceList(this);
		}
		void accept(Walker walker) {
			walker.visit(this);
		}
		public String toString() {
			return "(PrecedenceList " + super.toString() + ")";
		}
	}
	public static class Precedence extends Node {
		PrecedenceSymbolList precedenceSymbolList;
		Term assoc;
		public Precedence(PrecedenceSymbolList precedenceSymbolList, Term assoc) {
			(this.precedenceSymbolList = precedenceSymbolList).parent = this;
			(this.assoc = assoc).parent = this;
		}
		private Precedence(Precedence node) {
			(this.precedenceSymbolList = (PrecedenceSymbolList) node.precedenceSymbolList.dup()).parent = this;
			(this.assoc = (Term) node.assoc.dup()).parent = this;
		}
		Node dup() {
			return new Precedence(this);
		}
		void accept(Walker walker) {
			walker.visit(this);
		}
		public String toString() {
			return "(Precedence" + " " + precedenceSymbolList + " " + assoc + ")";
		}
	}
	public static class PrecedenceSymbolList extends NodeList {
		private PrecedenceSymbolList(PrecedenceSymbolList list) {
			super(list);
		}
		public PrecedenceSymbolList() {
			super();
		}
		public PrecedenceSymbolList(PrecedenceSymbol elem) {
			super(elem);
		}
		Node dup() {
			return new PrecedenceSymbolList(this);
		}
		void accept(Walker walker) {
			walker.visit(this);
		}
		public String toString() {
			return "(PrecedenceSymbolList " + super.toString() + ")";
		}
	}
	public static abstract class PrecedenceSymbol extends Node {
		public static class Name extends PrecedenceSymbol {
			Term name;
			public Name(Term name) {
				(this.name = name).parent = this;
			}
			private Name(Name node) {
				(this.name = (Term) node.name.dup()).parent = this;
			}
			Node dup() {
				return new Name(this);
			}
			void accept(Walker walker) {
				walker.visit(this);
			}
			public String toString() {
				return "(PrecedenceSymbol.Name" + " " + name + ")";
			}
		}
		public static class Text extends PrecedenceSymbol {
			Term text;
			public Text(Term text) {
				(this.text = text).parent = this;
			}
			private Text(Text node) {
				(this.text = (Term) node.text.dup()).parent = this;
			}
			Node dup() {
				return new Text(this);
			}
			void accept(Walker walker) {
				walker.visit(this);
			}
			public String toString() {
				return "(PrecedenceSymbol.Text" + " " + text + ")";
			}
		}
	}
	public static class MacroList extends NodeList {
		private MacroList(MacroList list) {
			super(list);
		}
		public MacroList() {
			super();
		}
		public MacroList(Macro elem) {
			super(elem);
		}
		Node dup() {
			return new MacroList(this);
		}
		void accept(Walker walker) {
			walker.visit(this);
		}
		public String toString() {
			return "(MacroList " + super.toString() + ")";
		}
	}
	public static class Macro extends Node {
		Term name;
		RegExp regExp;
		public Macro(Term name, RegExp regExp) {
			(this.name = name).parent = this;
			(this.regExp = regExp).parent = this;
		}
		private Macro(Macro node) {
			(this.name = (Term) node.name.dup()).parent = this;
			(this.regExp = (RegExp) node.regExp.dup()).parent = this;
		}
		Node dup() {
			return new Macro(this);
		}
		void accept(Walker walker) {
			walker.visit(this);
		}
		public String toString() {
			return "(Macro" + " " + name + " " + regExp + ")";
		}
	}
	public static abstract class RegExp extends Node {
		public static class Cat extends RegExp {
			RegExp regExp;
			RegExp regExp2;
			public Cat(RegExp regExp, RegExp regExp2) {
				(this.regExp = regExp).parent = this;
				(this.regExp2 = regExp2).parent = this;
			}
			private Cat(Cat node) {
				(this.regExp = (RegExp) node.regExp.dup()).parent = this;
				(this.regExp2 = (RegExp) node.regExp2.dup()).parent = this;
			}
			Node dup() {
				return new Cat(this);
			}
			void accept(Walker walker) {
				walker.visit(this);
			}
			public String toString() {
				return "(RegExp.Cat" + " " + regExp + " " + regExp2 + ")";
			}
		}
		public static class Alt extends RegExp {
			RegExp regExp;
			RegExp regExp2;
			public Alt(RegExp regExp, RegExp regExp2) {
				(this.regExp = regExp).parent = this;
				(this.regExp2 = regExp2).parent = this;
			}
			private Alt(Alt node) {
				(this.regExp = (RegExp) node.regExp.dup()).parent = this;
				(this.regExp2 = (RegExp) node.regExp2.dup()).parent = this;
			}
			Node dup() {
				return new Alt(this);
			}
			void accept(Walker walker) {
				walker.visit(this);
			}
			public String toString() {
				return "(RegExp.Alt" + " " + regExp + " " + regExp2 + ")";
			}
		}
		public static class Nested extends RegExp {
			RegExp regExp;
			Term optQuant;
			public Nested(RegExp regExp, Term optQuant) {
				(this.regExp = regExp).parent = this;
				if(optQuant != null) {
					(this.optQuant = optQuant).parent = this;
				}
			}
			private Nested(Nested node) {
				(this.regExp = (RegExp) node.regExp.dup()).parent = this;
				if(node.optQuant != null) {
					(this.optQuant = (Term) node.optQuant.dup()).parent = this;
				}
			}
			Node dup() {
				return new Nested(this);
			}
			void accept(Walker walker) {
				walker.visit(this);
			}
			public String toString() {
				return "(RegExp.Nested" + " " + regExp + asString(optQuant) + ")";
			}
		}
		public static class Macro extends RegExp {
			Term name;
			Term optQuant;
			public Macro(Term name, Term optQuant) {
				(this.name = name).parent = this;
				if(optQuant != null) {
					(this.optQuant = optQuant).parent = this;
				}
			}
			private Macro(Macro node) {
				(this.name = (Term) node.name.dup()).parent = this;
				if(node.optQuant != null) {
					(this.optQuant = (Term) node.optQuant.dup()).parent = this;
				}
			}
			Node dup() {
				return new Macro(this);
			}
			void accept(Walker walker) {
				walker.visit(this);
			}
			public String toString() {
				return "(RegExp.Macro" + " " + name + asString(optQuant) + ")";
			}
		}
		public static class Text extends RegExp {
			Term text;
			Term optQuant;
			public Text(Term text, Term optQuant) {
				(this.text = text).parent = this;
				if(optQuant != null) {
					(this.optQuant = optQuant).parent = this;
				}
			}
			private Text(Text node) {
				(this.text = (Term) node.text.dup()).parent = this;
				if(node.optQuant != null) {
					(this.optQuant = (Term) node.optQuant.dup()).parent = this;
				}
			}
			Node dup() {
				return new Text(this);
			}
			void accept(Walker walker) {
				walker.visit(this);
			}
			public String toString() {
				return "(RegExp.Text" + " " + text + asString(optQuant) + ")";
			}
		}
		public static class Range extends RegExp {
			Term range;
			Term optQuant;
			public Range(Term range, Term optQuant) {
				(this.range = range).parent = this;
				if(optQuant != null) {
					(this.optQuant = optQuant).parent = this;
				}
			}
			private Range(Range node) {
				(this.range = (Term) node.range.dup()).parent = this;
				if(node.optQuant != null) {
					(this.optQuant = (Term) node.optQuant.dup()).parent = this;
				}
			}
			Node dup() {
				return new Range(this);
			}
			void accept(Walker walker) {
				walker.visit(this);
			}
			public String toString() {
				return "(RegExp.Range" + " " + range + asString(optQuant) + ")";
			}
		}
		public static class SubRange extends RegExp {
			RangeExp rangeExp;
			RangeExp rangeExp2;
			public SubRange(RangeExp rangeExp, RangeExp rangeExp2) {
				(this.rangeExp = rangeExp).parent = this;
				(this.rangeExp2 = rangeExp2).parent = this;
			}
			private SubRange(SubRange node) {
				(this.rangeExp = (RangeExp) node.rangeExp.dup()).parent = this;
				(this.rangeExp2 = (RangeExp) node.rangeExp2.dup()).parent = this;
			}
			Node dup() {
				return new SubRange(this);
			}
			void accept(Walker walker) {
				walker.visit(this);
			}
			public String toString() {
				return "(RegExp.SubRange" + " " + rangeExp + " " + rangeExp2 + ")";
			}
		}
	}
	public static abstract class RangeExp extends Node {
		public static class Name extends RangeExp {
			Term name;
			public Name(Term name) {
				(this.name = name).parent = this;
			}
			private Name(Name node) {
				(this.name = (Term) node.name.dup()).parent = this;
			}
			Node dup() {
				return new Name(this);
			}
			void accept(Walker walker) {
				walker.visit(this);
			}
			public String toString() {
				return "(RangeExp.Name" + " " + name + ")";
			}
		}
		public static class Text extends RangeExp {
			Term range;
			public Text(Term range) {
				(this.range = range).parent = this;
			}
			private Text(Text node) {
				(this.range = (Term) node.range.dup()).parent = this;
			}
			Node dup() {
				return new Text(this);
			}
			void accept(Walker walker) {
				walker.visit(this);
			}
			public String toString() {
				return "(RangeExp.Text" + " " + range + ")";
			}
		}
	}
	public static class TokenList extends NodeList {
		private TokenList(TokenList list) {
			super(list);
		}
		public TokenList() {
			super();
		}
		public TokenList(Token elem) {
			super(elem);
		}
		Node dup() {
			return new TokenList(this);
		}
		void accept(Walker walker) {
			walker.visit(this);
		}
		public String toString() {
			return "(TokenList " + super.toString() + ")";
		}
	}
	public static class Token extends Node {
		Term name;
		StartCondList startCondList;
		RegExp regExp;
		EventList eventList;
		Term optCondShift;
		public Token(Term name, StartCondList startCondList, RegExp regExp, EventList eventList, Term optCondShift) {
			(this.name = name).parent = this;
			(this.startCondList = startCondList).parent = this;
			(this.regExp = regExp).parent = this;
			(this.eventList = eventList).parent = this;
			if(optCondShift != null) {
				(this.optCondShift = optCondShift).parent = this;
			}
		}
		private Token(Token node) {
			(this.name = (Term) node.name.dup()).parent = this;
			(this.startCondList = (StartCondList) node.startCondList.dup()).parent = this;
			(this.regExp = (RegExp) node.regExp.dup()).parent = this;
			(this.eventList = (EventList) node.eventList.dup()).parent = this;
			if(node.optCondShift != null) {
				(this.optCondShift = (Term) node.optCondShift.dup()).parent = this;
			}
		}
		Node dup() {
			return new Token(this);
		}
		void accept(Walker walker) {
			walker.visit(this);
		}
		public String toString() {
			return "(Token" + " " + name + " " + startCondList + " " + regExp + " " + eventList + asString(optCondShift) + ")";
		}
	}
	public static class StartCondList extends NodeList {
		private StartCondList(StartCondList list) {
			super(list);
		}
		public StartCondList() {
			super();
		}
		public StartCondList(Term elem) {
			super(elem);
		}
		Node dup() {
			return new StartCondList(this);
		}
		void accept(Walker walker) {
			walker.visit(this);
		}
		public String toString() {
			return "(StartCondList " + super.toString() + ")";
		}
	}
	public static class EventList extends NodeList {
		private EventList(EventList list) {
			super(list);
		}
		public EventList() {
			super();
		}
		public EventList(Term elem) {
			super(elem);
		}
		Node dup() {
			return new EventList(this);
		}
		void accept(Walker walker) {
			walker.visit(this);
		}
		public String toString() {
			return "(EventList " + super.toString() + ")";
		}
	}
	public static class Walker {
		void visit(Term node) {}
		boolean enter(Grammar node) {
			return true;
		}
		void leave(Grammar node) {}
		void visit(Grammar node) {
			if (enter(node)) {
				node.productionList.accept(this);
				node.precedenceList.accept(this);
				node.macroList.accept(this);
				node.tokenList.accept(this);
				leave(node);
			}
		}
		boolean enter(ProductionList node) {
			return true;
		}
		void leave(ProductionList node) {}
		void visit(ProductionList node) {
			if (enter(node)) {
				for (Node e = node.first; e != null; e = e.next) {
					e.accept(this);
				}
				leave(node);
			}
		}
		boolean enter(Production node) {
			return true;
		}
		void leave(Production node) {}
		void visit(Production node) {
			if (enter(node)) {
				visit(node.name);
				node.symbolDefinitionList.accept(this);
				leave(node);
			}
		}
		boolean enter(SymbolDefinitionList node) {
			return true;
		}
		void leave(SymbolDefinitionList node) {}
		void visit(SymbolDefinitionList node) {
			if (enter(node)) {
				for (Node e = node.first; e != null; e = e.next) {
					e.accept(this);
				}
				leave(node);
			}
		}
		boolean enter(SymbolDefinition node) {
			return true;
		}
		void leave(SymbolDefinition node) {}
		void visit(SymbolDefinition node) {
			if (enter(node)) {
				if (node.optRuleName != null) { node.optRuleName.accept(this); }
				node.symbolList.accept(this);
				leave(node);
			}
		}
		boolean enter(SymbolList node) {
			return true;
		}
		void leave(SymbolList node) {}
		void visit(SymbolList node) {
			if (enter(node)) {
				for (Node e = node.first; e != null; e = e.next) {
					e.accept(this);
				}
				leave(node);
			}
		}
		boolean enter(Symbol.Name node) {
			return true;
		}
		void leave(Symbol.Name node) {}
		void visit(Symbol.Name node) {
			if (enter(node)) {
				visit(node.name);
				if (node.optQuant != null) { node.optQuant.accept(this); }
				leave(node);
			}
		}
		boolean enter(Symbol.Text node) {
			return true;
		}
		void leave(Symbol.Text node) {}
		void visit(Symbol.Text node) {
			if (enter(node)) {
				visit(node.text);
				if (node.optQuant != null) { node.optQuant.accept(this); }
				leave(node);
			}
		}
		boolean enter(Symbol.Nested node) {
			return true;
		}
		void leave(Symbol.Nested node) {}
		void visit(Symbol.Nested node) {
			if (enter(node)) {
				node.symbolList.accept(this);
				if (node.optQuant != null) { node.optQuant.accept(this); }
				leave(node);
			}
		}
		boolean enter(PrecedenceList node) {
			return true;
		}
		void leave(PrecedenceList node) {}
		void visit(PrecedenceList node) {
			if (enter(node)) {
				for (Node e = node.first; e != null; e = e.next) {
					e.accept(this);
				}
				leave(node);
			}
		}
		boolean enter(Precedence node) {
			return true;
		}
		void leave(Precedence node) {}
		void visit(Precedence node) {
			if (enter(node)) {
				node.precedenceSymbolList.accept(this);
				visit(node.assoc);
				leave(node);
			}
		}
		boolean enter(PrecedenceSymbolList node) {
			return true;
		}
		void leave(PrecedenceSymbolList node) {}
		void visit(PrecedenceSymbolList node) {
			if (enter(node)) {
				for (Node e = node.first; e != null; e = e.next) {
					e.accept(this);
				}
				leave(node);
			}
		}
		boolean enter(PrecedenceSymbol.Name node) {
			return true;
		}
		void leave(PrecedenceSymbol.Name node) {}
		void visit(PrecedenceSymbol.Name node) {
			if (enter(node)) {
				visit(node.name);
				leave(node);
			}
		}
		boolean enter(PrecedenceSymbol.Text node) {
			return true;
		}
		void leave(PrecedenceSymbol.Text node) {}
		void visit(PrecedenceSymbol.Text node) {
			if (enter(node)) {
				visit(node.text);
				leave(node);
			}
		}
		boolean enter(MacroList node) {
			return true;
		}
		void leave(MacroList node) {}
		void visit(MacroList node) {
			if (enter(node)) {
				for (Node e = node.first; e != null; e = e.next) {
					e.accept(this);
				}
				leave(node);
			}
		}
		boolean enter(Macro node) {
			return true;
		}
		void leave(Macro node) {}
		void visit(Macro node) {
			if (enter(node)) {
				visit(node.name);
				node.regExp.accept(this);
				leave(node);
			}
		}
		boolean enter(RegExp.Cat node) {
			return true;
		}
		void leave(RegExp.Cat node) {}
		void visit(RegExp.Cat node) {
			if (enter(node)) {
				node.regExp.accept(this);
				node.regExp2.accept(this);
				leave(node);
			}
		}
		boolean enter(RegExp.Alt node) {
			return true;
		}
		void leave(RegExp.Alt node) {}
		void visit(RegExp.Alt node) {
			if (enter(node)) {
				node.regExp.accept(this);
				node.regExp2.accept(this);
				leave(node);
			}
		}
		boolean enter(RegExp.Nested node) {
			return true;
		}
		void leave(RegExp.Nested node) {}
		void visit(RegExp.Nested node) {
			if (enter(node)) {
				node.regExp.accept(this);
				if (node.optQuant != null) { node.optQuant.accept(this); }
				leave(node);
			}
		}
		boolean enter(RegExp.Macro node) {
			return true;
		}
		void leave(RegExp.Macro node) {}
		void visit(RegExp.Macro node) {
			if (enter(node)) {
				visit(node.name);
				if (node.optQuant != null) { node.optQuant.accept(this); }
				leave(node);
			}
		}
		boolean enter(RegExp.Text node) {
			return true;
		}
		void leave(RegExp.Text node) {}
		void visit(RegExp.Text node) {
			if (enter(node)) {
				visit(node.text);
				if (node.optQuant != null) { node.optQuant.accept(this); }
				leave(node);
			}
		}
		boolean enter(RegExp.Range node) {
			return true;
		}
		void leave(RegExp.Range node) {}
		void visit(RegExp.Range node) {
			if (enter(node)) {
				visit(node.range);
				if (node.optQuant != null) { node.optQuant.accept(this); }
				leave(node);
			}
		}
		boolean enter(RegExp.SubRange node) {
			return true;
		}
		void leave(RegExp.SubRange node) {}
		void visit(RegExp.SubRange node) {
			if (enter(node)) {
				node.rangeExp.accept(this);
				node.rangeExp2.accept(this);
				leave(node);
			}
		}
		boolean enter(RangeExp.Name node) {
			return true;
		}
		void leave(RangeExp.Name node) {}
		void visit(RangeExp.Name node) {
			if (enter(node)) {
				visit(node.name);
				leave(node);
			}
		}
		boolean enter(RangeExp.Text node) {
			return true;
		}
		void leave(RangeExp.Text node) {}
		void visit(RangeExp.Text node) {
			if (enter(node)) {
				visit(node.range);
				leave(node);
			}
		}
		boolean enter(TokenList node) {
			return true;
		}
		void leave(TokenList node) {}
		void visit(TokenList node) {
			if (enter(node)) {
				for (Node e = node.first; e != null; e = e.next) {
					e.accept(this);
				}
				leave(node);
			}
		}
		boolean enter(Token node) {
			return true;
		}
		void leave(Token node) {}
		void visit(Token node) {
			if (enter(node)) {
				visit(node.name);
				node.startCondList.accept(this);
				node.regExp.accept(this);
				node.eventList.accept(this);
				if (node.optCondShift != null) { node.optCondShift.accept(this); }
				leave(node);
			}
		}
		boolean enter(StartCondList node) {
			return true;
		}
		void leave(StartCondList node) {}
		void visit(StartCondList node) {
			if (enter(node)) {
				for (Node e = node.first; e != null; e = e.next) {
					e.accept(this);
				}
				leave(node);
			}
		}
		boolean enter(EventList node) {
			return true;
		}
		void leave(EventList node) {}
		void visit(EventList node) {
			if (enter(node)) {
				for (Node e = node.first; e != null; e = e.next) {
					e.accept(this);
				}
				leave(node);
			}
		}
	}
}
