package xml;
public abstract class AST {
	public static abstract class Node {
		Node parent;
		Node prev;
		Node next;
		abstract void accept(Walker walker);
		abstract Node dup();
		static String asString(Object object) {
			return object == null ? "" : " " + object.toString();
		}
	}
	public static class Term extends Node implements beaver.Term {
		static final String[] keywords = { "'<EOF>'", "'<?'", "'?>'", "'='", "'<'", "'/>'", "'>'", "'</'" };
		static final String[] tokens = { "COMMENT", "NAME", "VALUE", "TEXT" };
		int    id;
		int    line;
		int    column;
		Object text;
		public Term(int id, int line, int column, Object text) {
			this.id = id;
			this.line = line;
			this.column = column;
			this.text = text;
		}
		private Term(Term term) {
			this.id = term.id;
			this.line = term.line;
			this.column = term.column;
			this.text = term.text;
		}
		public void setId(int newId) {
			id = newId;
		}
		public int getId() {
			return id;
		}
		public Object getText() {
			return text;
		}
		public int getLine() {
			return line;
		}
		public int getColumn() {
			return column;
		}
		void accept(Walker walker) {
			walker.visit(this);
		}
		Node dup() {
			return new Term(this);
		}
		public String toString() {
			if (id >= 0) {
				if (id < keywords.length) {
					return keywords[id];
				}
				int ix = id - keywords.length;
				if (ix < tokens.length) {
					return tokens[ix];
				}
			}
			return text.toString();
		}
	}
	public static abstract class NodeList extends Node {
		Node first;
		Node last;
		int  size;
		NodeList() {}
		NodeList(Node node) {
			node.parent = this;
			last = first = node;
			size = 1;
		}
		protected NodeList(NodeList list) {
			Node elem = list.first;
			if (elem != null) {
				last = first = elem.dup();
				last.parent = this;
				last.prev = last.next = null;
				size = 1;
				for (elem = elem.next; elem != null; elem = elem.next) {
					add(elem.dup());
				}
			}
		}
		public void add(Node node) {
			node.parent = this;
			node.next = null;
			if ((node.prev = last) == null) {
				last = first = node;
			} else {
				last = last.next = node;
			}
			size++;
		}
		public void remove(Node node) {
			if (node.parent != this) {
				throw new IllegalArgumentException("wrong list");
			}
			if (node.prev != null) {
				first = node.next;
			} else {
				node.prev.next = node.next;
			}
			if (node.next == null) {
				last = node.prev;
			} else {
				node.next.prev = node.prev;
			}
			size--;
		}
		public void replace(Node node, Node newNode) {
			if (node.parent != this) {
				throw new IllegalArgumentException("wrong list");
			}
			newNode.parent = this;
			newNode.prev = node.prev;
			newNode.next = node.next;
			if (newNode.prev == null) {
				first = newNode;
			} else {
				newNode.prev.next = newNode;
			}
			if (newNode.next == null) {
				last = newNode;
			} else {
				newNode.next.prev = newNode;
			}
		}
		public String toString() {
			if (first != null) {
				String str = first.toString();
				for (Node e = first.next; e != null; e = e.next) {
					str += ", " + e;
				}
				return str;
			}
			return "";
		}
	}
	public static class Document extends Node {
		OptMiscElemList optMiscElemList;
		Element element;
		OptMiscElemList optMiscElemList2;
		public Document(OptMiscElemList optMiscElemList, Element element, OptMiscElemList optMiscElemList2) {
			(this.optMiscElemList = optMiscElemList).parent = this;
			(this.element = element).parent = this;
			(this.optMiscElemList2 = optMiscElemList2).parent = this;
		}
		private Document(Document node) {
			(this.optMiscElemList = (OptMiscElemList) node.optMiscElemList.dup()).parent = this;
			(this.element = (Element) node.element.dup()).parent = this;
			(this.optMiscElemList2 = (OptMiscElemList) node.optMiscElemList2.dup()).parent = this;
		}
		Node dup() {
			return new Document(this);
		}
		void accept(Walker walker) {
			walker.visit(this);
		}
		public String toString() {
			return "(Document" + " " + optMiscElemList + " " + element + " " + optMiscElemList2 + ")";
		}
	}
	public static class OptMiscElemList extends NodeList {
		private OptMiscElemList(OptMiscElemList list) {
			super(list);
		}
		public OptMiscElemList() {
			super();
		}
		public OptMiscElemList(MiscElem elem) {
			super(elem);
		}
		Node dup() {
			return new OptMiscElemList(this);
		}
		void accept(Walker walker) {
			walker.visit(this);
		}
		public String toString() {
			return "(OptMiscElemList " + super.toString() + ")";
		}
	}
	public static abstract class MiscElem extends Node {
		public static class Comment extends MiscElem {
			Term comment;
			public Comment(Term comment) {
				(this.comment = comment).parent = this;
			}
			private Comment(Comment node) {
				(this.comment = (Term) node.comment.dup()).parent = this;
			}
			Node dup() {
				return new Comment(this);
			}
			void accept(Walker walker) {
				walker.visit(this);
			}
			public String toString() {
				return "(MiscElem.Comment" + " " + comment + ")";
			}
		}
		public static class Instr extends MiscElem {
			ProcInstr procInstr;
			public Instr(ProcInstr procInstr) {
				(this.procInstr = procInstr).parent = this;
			}
			private Instr(Instr node) {
				(this.procInstr = (ProcInstr) node.procInstr.dup()).parent = this;
			}
			Node dup() {
				return new Instr(this);
			}
			void accept(Walker walker) {
				walker.visit(this);
			}
			public String toString() {
				return "(MiscElem.Instr" + " " + procInstr + ")";
			}
		}
	}
	public static class ProcInstr extends Node {
		Term name;
		OptAttrList optAttrList;
		public ProcInstr(Term name, OptAttrList optAttrList) {
			(this.name = name).parent = this;
			(this.optAttrList = optAttrList).parent = this;
		}
		private ProcInstr(ProcInstr node) {
			(this.name = (Term) node.name.dup()).parent = this;
			(this.optAttrList = (OptAttrList) node.optAttrList.dup()).parent = this;
		}
		Node dup() {
			return new ProcInstr(this);
		}
		void accept(Walker walker) {
			walker.visit(this);
		}
		public String toString() {
			return "(ProcInstr" + " " + name + " " + optAttrList + ")";
		}
	}
	public static class OptAttrList extends NodeList {
		private OptAttrList(OptAttrList list) {
			super(list);
		}
		public OptAttrList() {
			super();
		}
		public OptAttrList(Attribute elem) {
			super(elem);
		}
		Node dup() {
			return new OptAttrList(this);
		}
		void accept(Walker walker) {
			walker.visit(this);
		}
		public String toString() {
			return "(OptAttrList " + super.toString() + ")";
		}
	}
	public static abstract class Attribute extends Node {
		public static class NameOnly extends Attribute {
			Term name;
			public NameOnly(Term name) {
				(this.name = name).parent = this;
			}
			private NameOnly(NameOnly node) {
				(this.name = (Term) node.name.dup()).parent = this;
			}
			Node dup() {
				return new NameOnly(this);
			}
			void accept(Walker walker) {
				walker.visit(this);
			}
			public String toString() {
				return "(Attribute.NameOnly" + " " + name + ")";
			}
		}
		public static class NameAndValue extends Attribute {
			Term name;
			Term value;
			public NameAndValue(Term name, Term value) {
				(this.name = name).parent = this;
				(this.value = value).parent = this;
			}
			private NameAndValue(NameAndValue node) {
				(this.name = (Term) node.name.dup()).parent = this;
				(this.value = (Term) node.value.dup()).parent = this;
			}
			Node dup() {
				return new NameAndValue(this);
			}
			void accept(Walker walker) {
				walker.visit(this);
			}
			public String toString() {
				return "(Attribute.NameAndValue" + " " + name + " " + value + ")";
			}
		}
	}
	public static abstract class Element extends Node {
		public static class Standalone extends Element {
			Term name;
			OptAttrList optAttrList;
			public Standalone(Term name, OptAttrList optAttrList) {
				(this.name = name).parent = this;
				(this.optAttrList = optAttrList).parent = this;
			}
			private Standalone(Standalone node) {
				(this.name = (Term) node.name.dup()).parent = this;
				(this.optAttrList = (OptAttrList) node.optAttrList.dup()).parent = this;
			}
			Node dup() {
				return new Standalone(this);
			}
			void accept(Walker walker) {
				walker.visit(this);
			}
			public String toString() {
				return "(Element.Standalone" + " " + name + " " + optAttrList + ")";
			}
		}
		public static class WithContent extends Element {
			Term name;
			OptAttrList optAttrList;
			Content content;
			Term name2;
			public WithContent(Term name, OptAttrList optAttrList, Content content, Term name2) {
				(this.name = name).parent = this;
				(this.optAttrList = optAttrList).parent = this;
				(this.content = content).parent = this;
				(this.name2 = name2).parent = this;
			}
			private WithContent(WithContent node) {
				(this.name = (Term) node.name.dup()).parent = this;
				(this.optAttrList = (OptAttrList) node.optAttrList.dup()).parent = this;
				(this.content = (Content) node.content.dup()).parent = this;
				(this.name2 = (Term) node.name2.dup()).parent = this;
			}
			Node dup() {
				return new WithContent(this);
			}
			void accept(Walker walker) {
				walker.visit(this);
			}
			public String toString() {
				return "(Element.WithContent" + " " + name + " " + optAttrList + " " + content + " " + name2 + ")";
			}
		}
	}
	public static class Content extends NodeList {
		private Content(Content list) {
			super(list);
		}
		public Content() {
			super();
		}
		public Content(ContentItem elem) {
			super(elem);
		}
		Node dup() {
			return new Content(this);
		}
		void accept(Walker walker) {
			walker.visit(this);
		}
		public String toString() {
			return "(Content " + super.toString() + ")";
		}
	}
	public static abstract class ContentItem extends Node {
		public static class Text extends ContentItem {
			Term text;
			public Text(Term text) {
				(this.text = text).parent = this;
			}
			private Text(Text node) {
				(this.text = (Term) node.text.dup()).parent = this;
			}
			Node dup() {
				return new Text(this);
			}
			void accept(Walker walker) {
				walker.visit(this);
			}
			public String toString() {
				return "(ContentItem.Text" + " " + text + ")";
			}
		}
		public static class Misc extends ContentItem {
			MiscElem miscElem;
			public Misc(MiscElem miscElem) {
				(this.miscElem = miscElem).parent = this;
			}
			private Misc(Misc node) {
				(this.miscElem = (MiscElem) node.miscElem.dup()).parent = this;
			}
			Node dup() {
				return new Misc(this);
			}
			void accept(Walker walker) {
				walker.visit(this);
			}
			public String toString() {
				return "(ContentItem.Misc" + " " + miscElem + ")";
			}
		}
		public static class Elem extends ContentItem {
			Element element;
			public Elem(Element element) {
				(this.element = element).parent = this;
			}
			private Elem(Elem node) {
				(this.element = (Element) node.element.dup()).parent = this;
			}
			Node dup() {
				return new Elem(this);
			}
			void accept(Walker walker) {
				walker.visit(this);
			}
			public String toString() {
				return "(ContentItem.Elem" + " " + element + ")";
			}
		}
	}
	public static class Walker {
		void visit(Term node) {}
		boolean enter(Document node) {
			return true;
		}
		void leave(Document node) {}
		void visit(Document node) {
			if (enter(node)) {
				node.optMiscElemList.accept(this);
				node.element.accept(this);
				node.optMiscElemList2.accept(this);
				leave(node);
			}
		}
		boolean enter(OptMiscElemList node) {
			return true;
		}
		void leave(OptMiscElemList node) {}
		void visit(OptMiscElemList node) {
			if (enter(node)) {
				for (Node e = node.first; e != null; e = e.next) {
					e.accept(this);
				}
				leave(node);
			}
		}
		boolean enter(MiscElem.Comment node) {
			return true;
		}
		void leave(MiscElem.Comment node) {}
		void visit(MiscElem.Comment node) {
			if (enter(node)) {
				visit(node.comment);
				leave(node);
			}
		}
		boolean enter(MiscElem.Instr node) {
			return true;
		}
		void leave(MiscElem.Instr node) {}
		void visit(MiscElem.Instr node) {
			if (enter(node)) {
				node.procInstr.accept(this);
				leave(node);
			}
		}
		boolean enter(ProcInstr node) {
			return true;
		}
		void leave(ProcInstr node) {}
		void visit(ProcInstr node) {
			if (enter(node)) {
				visit(node.name);
				node.optAttrList.accept(this);
				leave(node);
			}
		}
		boolean enter(OptAttrList node) {
			return true;
		}
		void leave(OptAttrList node) {}
		void visit(OptAttrList node) {
			if (enter(node)) {
				for (Node e = node.first; e != null; e = e.next) {
					e.accept(this);
				}
				leave(node);
			}
		}
		boolean enter(Attribute.NameOnly node) {
			return true;
		}
		void leave(Attribute.NameOnly node) {}
		void visit(Attribute.NameOnly node) {
			if (enter(node)) {
				visit(node.name);
				leave(node);
			}
		}
		boolean enter(Attribute.NameAndValue node) {
			return true;
		}
		void leave(Attribute.NameAndValue node) {}
		void visit(Attribute.NameAndValue node) {
			if (enter(node)) {
				visit(node.name);
				visit(node.value);
				leave(node);
			}
		}
		boolean enter(Element.Standalone node) {
			return true;
		}
		void leave(Element.Standalone node) {}
		void visit(Element.Standalone node) {
			if (enter(node)) {
				visit(node.name);
				node.optAttrList.accept(this);
				leave(node);
			}
		}
		boolean enter(Element.WithContent node) {
			return true;
		}
		void leave(Element.WithContent node) {}
		void visit(Element.WithContent node) {
			if (enter(node)) {
				visit(node.name);
				node.optAttrList.accept(this);
				node.content.accept(this);
				visit(node.name2);
				leave(node);
			}
		}
		boolean enter(Content node) {
			return true;
		}
		void leave(Content node) {}
		void visit(Content node) {
			if (enter(node)) {
				for (Node e = node.first; e != null; e = e.next) {
					e.accept(this);
				}
				leave(node);
			}
		}
		boolean enter(ContentItem.Text node) {
			return true;
		}
		void leave(ContentItem.Text node) {}
		void visit(ContentItem.Text node) {
			if (enter(node)) {
				visit(node.text);
				leave(node);
			}
		}
		boolean enter(ContentItem.Misc node) {
			return true;
		}
		void leave(ContentItem.Misc node) {}
		void visit(ContentItem.Misc node) {
			if (enter(node)) {
				node.miscElem.accept(this);
				leave(node);
			}
		}
		boolean enter(ContentItem.Elem node) {
			return true;
		}
		void leave(ContentItem.Elem node) {}
		void visit(ContentItem.Elem node) {
			if (enter(node)) {
				node.element.accept(this);
				leave(node);
			}
		}
	}
}
